#!/usr/bin/python

"""
_SimpleCondorPlugin_t_

SimpleCondorPlugin unittests
"""
from __future__ import division, print_function

from future.utils import viewitems

import os.path
import re
import threading
import time
import unittest
from subprocess import Popen, PIPE

from WMCore_t.BossAir_t.BossAir_t import BossAirTest, getCondorRunningJobs
from nose.plugins.attrib import attr

from WMComponent.JobSubmitter.JobSubmitterPoller import JobSubmitterPoller
from WMComponent.JobTracker.JobTrackerPoller import JobTrackerPoller
from WMCore.BossAir.BossAirAPI import BossAirAPI
from WMCore.BossAir.StatusPoller import StatusPoller
from WMCore.BossAir.Plugins.SimpleCondorPlugin import activityToType
from WMCore.JobStateMachine.ChangeState import ChangeState


class SimpleCondorPluginTest(BossAirTest):
    """
    _SimpleCondorPluginTest_

    Inherit everything from BossAir
    """

    @attr('integration')
    def testC_CondorTest(self):
        """
        _CondorTest_

        This test works on the SimpleCondorPlugin, checking all of
        its functions with a single set of jobs
        """
        nRunning = getCondorRunningJobs(self.user)
        self.assertEqual(nRunning, 0, "User currently has %i running jobs.  Test will not continue" % (nRunning))

        # Get the config and set the removal time to -10 for testing
        config = self.getConfig()
        config.BossAir.removeTime = -10.0

        nJobs = 10
        jobDummies = self.createDummyJobs(nJobs=nJobs)
        baAPI = BossAirAPI(config=config, insertStates=True)

        jobPackage = os.path.join(self.testDir, 'JobPackage.pkl')
        f = open(jobPackage, 'w')
        f.write(' ')
        f.close()

        sandbox = os.path.join(self.testDir, 'sandbox.box')
        f = open(sandbox, 'w')
        f.write(' ')
        f.close()

        jobList = []
        for j in jobDummies:
            tmpJob = {'id': j['id']}
            tmpJob['custom'] = {'location': 'malpaquet'}
            tmpJob['name'] = j['name']
            tmpJob['cache_dir'] = self.testDir
            tmpJob['retry_count'] = 0
            tmpJob['plugin'] = 'SimpleCondorPlugin'
            tmpJob['owner'] = 'tapas'
            tmpJob['packageDir'] = self.testDir
            tmpJob['sandbox'] = sandbox
            tmpJob['priority'] = None
            tmpJob['usergroup'] = "wheel"
            tmpJob['userrole'] = 'cmsuser'
            jobList.append(tmpJob)

        info = {}
        # info['packageDir'] = self.testDir
        info['index'] = 0
        info['sandbox'] = sandbox

        baAPI.submit(jobs=jobList, info=info)

        nRunning = getCondorRunningJobs(self.user)
        self.assertEqual(nRunning, nJobs)

        newJobs = baAPI._loadByStatus(status='New')
        self.assertEqual(len(newJobs), nJobs)

        baAPI.track()

        newJobs = baAPI._loadByStatus(status='New')
        self.assertEqual(len(newJobs), 0)

        newJobs = baAPI._loadByStatus(status='Idle')
        self.assertEqual(len(newJobs), nJobs)

        # Do a second time to make sure that the cache
        # doesn't die on us
        baAPI.track()

        newJobs = baAPI._loadByStatus(status='New')
        self.assertEqual(len(newJobs), 0)

        newJobs = baAPI._loadByStatus(status='Idle')
        self.assertEqual(len(newJobs), nJobs)

        baAPI.kill(jobs=jobList)

        nRunning = getCondorRunningJobs(self.user)
        self.assertEqual(nRunning, 0)

        # Try resubmission
        for j in jobList:
            j['retry_count'] = 1

        baAPI.submit(jobs=jobList, info=info)

        nRunning = getCondorRunningJobs(self.user)
        self.assertEqual(nRunning, nJobs)

        newJobs = baAPI._loadByStatus(status='New')
        self.assertEqual(len(newJobs), nJobs)

        # See where they are
        baAPI.track()

        newJobs = baAPI._loadByStatus(status='New')
        self.assertEqual(len(newJobs), 0)

        newJobs = baAPI._loadByStatus(status='Idle')
        self.assertEqual(len(newJobs), nJobs)

        # Now kill 'em manually
        command = ['condor_rm', self.user]
        pipe = Popen(command, stdout=PIPE, stderr=PIPE, shell=False)
        pipe.communicate()

        # See what happened
        baAPI.track()

        newJobs = baAPI._loadByStatus(status='Idle')
        self.assertEqual(len(newJobs), 0)

        # newJobs = baAPI._loadByStatus(status = 'Removed')
        # self.assertEqual(len(newJobs), nJobs)

        # Because removal time is -10.0, jobs should remove immediately
        baAPI.track()

        # Assert that jobs were listed as completed
        myThread = threading.currentThread()
        newJobs = baAPI._loadByStatus(status='Removed', complete='0')
        self.assertEqual(len(newJobs), nJobs)
        return

    @attr('integration')
    def testD_PrototypeChain(self):
        """
        _PrototypeChain_

        Prototype the BossAir workflow
        """
        dummymyThread = threading.currentThread()

        nRunning = getCondorRunningJobs(self.user)
        self.assertEqual(nRunning, 0, "User currently has %i running jobs.  Test will not continue" % (nRunning))

        config = self.getConfig()
        config.BossAir.pluginName = 'SimpleCondorPlugin'

        baAPI = BossAirAPI(config=config, insertStates=True)

        workload = self.createTestWorkload()

        workloadName = "basicWorkload"

        changeState = ChangeState(config)

        nSubs = 5
        nJobs = 10

        dummycacheDir = os.path.join(self.testDir, 'CacheDir')

        jobGroupList = self.createJobGroups(nSubs=nSubs, nJobs=nJobs,
                                            task=workload.getTask("ReReco"),
                                            workloadSpec=os.path.join(self.testDir,
                                                                      'workloadTest',
                                                                      workloadName),
                                            site='se.T2_US_UCSD')
        for group in jobGroupList:
            changeState.propagate(group.jobs, 'created', 'new')

        jobSubmitter = JobSubmitterPoller(config=config)
        jobTracker = JobTrackerPoller(config=config)
        statusPoller = StatusPoller(config=config)

        jobSubmitter.algorithm()

        nRunning = getCondorRunningJobs(self.user)
        self.assertEqual(nRunning, nSubs * nJobs)

        newJobs = baAPI._loadByStatus(status='New')
        self.assertEqual(len(newJobs), nSubs * nJobs)

        # Check WMBS
        getJobsAction = self.daoFactory(classname="Jobs.GetAllJobs")
        result = getJobsAction.execute(state='Executing', jobType="Processing")
        self.assertEqual(len(result), nSubs * nJobs)

        statusPoller.algorithm()

        nRunning = getCondorRunningJobs(self.user)
        self.assertEqual(nRunning, nSubs * nJobs)

        newJobs = baAPI._loadByStatus(status='New')
        self.assertEqual(len(newJobs), 0)

        newJobs = baAPI._loadByStatus(status='Idle')
        self.assertEqual(len(newJobs), nSubs * nJobs)

        # Tracker should do nothing
        jobTracker.algorithm()

        result = getJobsAction.execute(state='Executing', jobType="Processing")
        self.assertEqual(len(result), nSubs * nJobs)

        # Wait for jobs to timeout due to short Pending wait period
        time.sleep(12)

        statusPoller.algorithm()

        newJobs = baAPI._loadByStatus(status='Idle')
        self.assertEqual(len(newJobs), 0)

        newJobs = baAPI._loadByStatus(status='Timeout', complete='0')
        self.assertEqual(len(newJobs), nSubs * nJobs)

        # Jobs should be gone
        nRunning = getCondorRunningJobs(self.user)
        self.assertEqual(nRunning, 0)

        # Check if they're complete
        completeJobs = baAPI.getComplete()
        self.assertEqual(len(completeJobs), nSubs * nJobs)

        # Because they timed out, they all should have failed
        jobTracker.algorithm()

        result = getJobsAction.execute(state='Executing', jobType="Processing")
        self.assertEqual(len(result), 0)

        result = getJobsAction.execute(state='JobFailed', jobType="Processing")
        self.assertEqual(len(result), nSubs * nJobs)

        return

    @attr('integration')
    def testE_FullChain(self):
        """
        _FullChain_

        Full test going through the chain; using polling cycles and everything
        """

        from WMComponent.JobSubmitter.JobSubmitter import JobSubmitter
        from WMComponent.JobStatusLite.JobStatusLite import JobStatusLite
        from WMComponent.JobTracker.JobTracker import JobTracker

        myThread = threading.currentThread()

        nRunning = getCondorRunningJobs(self.user)
        self.assertEqual(nRunning, 0, "User currently has %i running jobs.  Test will not continue" % (nRunning))

        config = self.getConfig()
        config.BossAir.pluginName = 'SimpleCondorPlugin'

        baAPI = BossAirAPI(config=config, insertStates=True)

        workload = self.createTestWorkload()

        workloadName = "basicWorkload"

        changeState = ChangeState(config)

        nSubs = 1
        nJobs = 2
        cacheDir = os.path.join(self.testDir, 'CacheDir')

        jobGroupList = self.createJobGroups(nSubs=nSubs, nJobs=nJobs,
                                            task=workload.getTask("ReReco"),
                                            workloadSpec=os.path.join(self.testDir,
                                                                      'workloadTest',
                                                                      workloadName),
                                            site='se.T2_US_UCSD')
        for group in jobGroupList:
            changeState.propagate(group.jobs, 'created', 'new')

        jobSubmitter = JobSubmitter(config=config)
        jobTracker = JobTracker(config=config)
        jobStatus = JobStatusLite(config=config)

        jobSubmitter.prepareToStart()
        jobTracker.prepareToStart()
        jobStatus.prepareToStart()

        # What should happen here:
        # 1) The JobSubmitter should submit the jobs
        # 2) Because of the ridiculously short time on pending jobs
        #     the JobStatus poller should mark the jobs as done
        #     and kill them.
        # 3) The JobTracker should realize there are finished jobs
        #
        # So at the end of several polling cycles, the jobs should all
        # be done, but be in the failed status (they timed out)

        time.sleep(20)

        myThread.workerThreadManager.terminateWorkers()

        getJobsAction = self.daoFactory(classname="Jobs.GetAllJobs")
        result = getJobsAction.execute(state='Executing', jobType="Processing")
        self.assertEqual(len(result), 0)

        result = getJobsAction.execute(state='JobFailed', jobType="Processing")
        self.assertEqual(len(result), nJobs * nSubs)
        return

    @attr('integration')
    def testF_WMSMode(self):
        """
        _WMSMode_

        Try running things in WMS Mode.
        """
        nRunning = getCondorRunningJobs(self.user)
        self.assertEqual(nRunning, 0, "User currently has %i running jobs.  Test will not continue" % (nRunning))

        config = self.getConfig()
        config.BossAir.pluginName = 'SimpleCondorPlugin'
        config.BossAir.submitWMSMode = True

        baAPI = BossAirAPI(config=config, insertStates=True)

        workload = self.createTestWorkload()

        workloadName = "basicWorkload"

        changeState = ChangeState(config)

        nSubs = 5
        nJobs = 10

        dummycacheDir = os.path.join(self.testDir, 'CacheDir')

        jobGroupList = self.createJobGroups(nSubs=nSubs, nJobs=nJobs,
                                            task=workload.getTask("ReReco"),
                                            workloadSpec=os.path.join(self.testDir,
                                                                      'workloadTest',
                                                                      workloadName),
                                            site=None)
        for group in jobGroupList:
            changeState.propagate(group.jobs, 'created', 'new')

        jobSubmitter = JobSubmitterPoller(config=config)

        jobSubmitter.algorithm()

        nRunning = getCondorRunningJobs(self.user)
        self.assertEqual(nRunning, nSubs * nJobs)

        baAPI.track()
        idleJobs = baAPI._loadByStatus(status='Idle')

        baAPI.kill(jobs=idleJobs)

        del jobSubmitter

        return

    @attr('integration')
    def testT_updateJobInfo(self):
        """
        _updateJobInfo_

        Test the updateSiteInformation method from SimpleCondorPlugin.py
        """
        nRunning = getCondorRunningJobs(self.user)
        self.assertEqual(nRunning, 0, "User currently has %i running jobs.  Test will not continue" % (nRunning))

        config = self.getConfig()
        config.BossAir.pluginName = 'SimpleCondorPlugin'
        config.BossAir.submitWMSMode = True

        baAPI = BossAirAPI(config=config, insertStates=True)
        workload = self.createTestWorkload()
        workloadName = "basicWorkload"
        changeState = ChangeState(config)

        nSubs = 1
        nJobs = 2
        dummycacheDir = os.path.join(self.testDir, 'CacheDir')
        jobGroupList = self.createJobGroups(nSubs=nSubs, nJobs=nJobs,
                                            task=workload.getTask("ReReco"),
                                            workloadSpec=os.path.join(self.testDir,
                                                                      'workloadTest',
                                                                      workloadName),
                                            site="se.T2_US_UCSD")
        for group in jobGroupList:
            changeState.propagate(group.jobs, 'created', 'new')
        jobSubmitter = JobSubmitterPoller(config=config)
        jobSubmitter.algorithm()
        nRunning = getCondorRunningJobs(self.user)
        self.assertEqual(nRunning, nSubs * nJobs)

        baAPI.track()
        idleJobs = baAPI._loadByStatus(status='Idle')

        # #
        # Make one of the sites in the sitelist to be True for ABORTED/DRAINING/DOWN
        # updateSiteInformation() method should edit the classAd for all the jobs
        # that are bound for the site
        # Check the Q manually using condor_q -l <job id>
        #
        jtok = baAPI.updateSiteInformation(idleJobs, "T2_US_UCSD", True)
        if jtok is not None:
            baAPI.kill(jtok, errorCode=71301)  # errorCode can be either 71301/71302/71303 (Aborted/Draining/Down)

        return

    def test_CMSGroupsRegex(self):
        """
        _test_CMSGroupsRegex_

        Simply test the regular expression for the CMSGroups classad
        """
        GROUP_NAME_RE = re.compile("^[a-zA-Z0-9_]+_([a-zA-Z0-9]+)-")

        mapGroupReqName = {'HIRun2015': 'jen_a_ACDC_HIRun2015-HIMinimumBias5-02May2016_758p4_160721_015926_2101',
                           'HIG': 'pdmvserv_HIG-RunIISummer15GS-01331_00274_v2__160308_111346_296',
                           'EXO': 'pdmvserv_task_EXO-RunIISpring16DR80-03516__v1_T_160719_140341_8564',
                           'BPH': 'pdmvserv_BPH-Summer12-00208_00263_v1__160726_025423_731',
                           'undefined': 'amaltaro_TaskChain_MC_DMWM_Test_July_Patches_160727_234956_6253'}

        for group, request in viewitems(mapGroupReqName):
            match = GROUP_NAME_RE.match(request)
            matchedGroup = match.groups()[0] if match else 'undefined'
            self.assertEqual(group, matchedGroup)

    def testActivityToTypeMap(self):
        """
        _testActivityToTypeMap_
        Test mapping from dashboard activity to cms type
        """
        self.assertEqual(activityToType("production"), "production")
        self.assertEqual(activityToType("reprocessing"), "production")
        self.assertEqual(activityToType("relval"), "production")
        self.assertEqual(activityToType("tier0"), "tier0")
        self.assertEqual(activityToType("integration"), "test")
        self.assertEqual(activityToType("alan"), "unknown")
        self.assertEqual(activityToType("none"), "unknown")
        self.assertEqual(activityToType(None), "unknown")



if __name__ == '__main__':
    unittest.main()



from builtins import range
import unittest
import os
import sys
from datetime import timedelta
from time import sleep

from nose.plugins.attrib import attr

from WMCore.Database.CMSCouch import RotatingDatabase, CouchServer, CouchNotFoundError


class RotatingDatabaseTest(unittest.TestCase):
    def setUp(self):
        self.couchURL = os.getenv("COUCHURL")
        self.server = CouchServer(self.couchURL)

        testname = self.id().split('.')[-1].lower()
        self.dbname = 'rotdb_unittest_%s' % testname
        self.arcname = 'rotdb_unittest_%s_archive' % testname
        self.seedname = 'rotdb_unittest_%s_seedcfg' % testname
        # set a long value for times, tests do operations explicitly
        self.timing = {'archive':timedelta(seconds=1), 'expire':timedelta(seconds=2)}

        self.db = RotatingDatabase(dbname = self.dbname, url = self.couchURL,
                                   archivename = self.arcname, timing = self.timing)

    def tearDown(self):
        """Delete all the test couchdb databases"""
        to_go = [db for db in self.server.listDatabases() if db.startswith('rotdb_unittest_')]
        for dbname in to_go:
            try:
                self.server.deleteDatabase(dbname)
            except CouchNotFoundError:
                # db has already gone
                pass

    def testRotate(self):
        """
        Test that rotation works
        """
        start_name = self.db.name
        self.db._rotate()
        end_name = self.db.name
        databases = [db for db in self.server.listDatabases() if db.startswith('rotdb_unittest_')]
        self.assertTrue(start_name in databases)
        self.assertTrue(end_name in databases)

    # This test repeatably inserts either 20 or 25 documents into couch.
    # Disabled until it's stable
    @attr("integration")
    def testArchive(self):
        """
        Test that archiving views works
        """
        dummy_view = {'_id':'_design/foo', 'language': 'javascript','views':{
                        'bar':{'map':"function(doc) {if (doc.foo) {emit(doc.int, 1);}}", 'reduce':'_sum'}
                        }
                    }
        archive_view = {'_id':'_design/foo', 'language': 'javascript','views':{
                        'bar':{'map':"function(doc) {emit(doc.key, doc.value);}", 'reduce':'_sum'}
                        }
                    }

        seed_db = self.server.connectDatabase(self.seedname)
        seed_db.commit(dummy_view)
        # Need to have the timing long enough so the data isn't archived by accident
        self.timing = {'archive':timedelta(seconds=1000), 'expire':timedelta(seconds=2000)}
        self.db = RotatingDatabase(dbname = self.dbname, url = self.couchURL, views=['foo/bar'],
                                archivename = self.arcname, timing = self.timing)
        self.db.archive_db.commitOne(archive_view)
        runs = 5
        docs = 5
        for run in range(runs):
            for i in range(docs):
                self.db.queue({'foo':'bar', 'int': i, 'run': run})
            self.db.commit()
            self.db._rotate()
        self.db._archive()
        view_result = self.db.archive_db.loadView('foo','bar')
        arch_sum = view_result['rows'][0]['value']
        self.assertEqual(arch_sum, runs * docs)

    def testExpire(self):
        """
        Test that expiring databases works
        """
        # rotate out the original db
        self.db._rotate()
        archived = self.db.archived_dbs()
        self.assertEqual(1, len(archived), 'test not starting from clean state, bail!')
        # Make sure the db has expired
        sleep(2)
        self.db._expire()
        self.assertEqual(0, len(self.db.archived_dbs()))
        self.assertFalse(archived[0] in self.server.listDatabases())

    @attr("integration")
    def testCycle(self):
        """
        Test that committing data to different databases happens
        This is a bit of a dodgy test - if timings go funny it will fail
        """
        self.timing = {'archive':timedelta(seconds=0.5), 'expire':timedelta(seconds=1)}
        self.db = RotatingDatabase(dbname = self.dbname, url = self.couchURL,
                                   archivename = self.arcname, timing = self.timing)
        my_name = self.db.name
        self.db.commit({'foo':'bar'})
        sleep(5)
        self.db.commit({'foo':'bar'})
        # the initial db should have expired by now
        self.db.commit({'foo':'bar'})
        self.assertFalse(my_name in self.server.listDatabases(), "")


if __name__ == "__main__":
    if len(sys.argv) >1 :
        suite = unittest.TestSuite()
        suite.addTest(RotatingDatabaseTest(sys.argv[1]))
        unittest.TextTestRunner().run(suite)
    else:
        unittest.main()



#!/usr/bin/env python
"""
Module for finding out workflows with open subscriptions
for production and/or processing tasks.
"""

from WMCore.Database.DBFormatter import DBFormatter


class GetUnfinishedWorkflows(DBFormatter):
    """
    If a workflow has any production or processing type subscriptions
    still unfinished, then it is considered an unfinished/active workflow.
    """

    sql = """SELECT wmbs_workflow.name, wmbs_workflow.spec
               FROM wmbs_workflow
               INNER JOIN wmbs_subscription ON wmbs_subscription.workflow = wmbs_workflow.id
               INNER JOIN wmbs_sub_types ON wmbs_sub_types.id = wmbs_subscription.subtype
               WHERE wmbs_subscription.finished = 0
                 AND wmbs_sub_types.name IN ('Processing', 'Production') 
               GROUP BY wmbs_workflow.name, wmbs_workflow.spec"""

    def execute(self, conn=None, transaction=False):
        """
        Executes the SQL statement above.
        :param conn: connection object to the database
        :param transaction: boolean defining whether it's part of a transaction or not
        :return: a list of dictionary elements
        """
        result = self.dbi.processData(self.sql, conn=conn, transaction=transaction)
        return self.formatDict(result)



#!/usr/bin/env python
"""
The JobCreator Poller for the JSM
"""

from builtins import next

__all__ = []

import logging
import os
import os.path
import threading
import pickle

from Utils.Timers import timeFunction
from Utils.PythonVersion import HIGHEST_PICKLE_PROTOCOL
from Utils.MathUtils import quantize
from WMComponent.JobCreator.CreateWorkArea import CreateWorkArea
from WMCore.WorkerThreads.BaseWorkerThread import BaseWorkerThread
from WMCore.DAOFactory import DAOFactory
from WMCore.WMException import WMException
from WMCore.JobSplitting.Generators.GeneratorManager import GeneratorManager
from WMCore.JobStateMachine.ChangeState import ChangeState
from WMCore.JobSplitting.SplitterFactory import SplitterFactory
from WMCore.WMBS.Subscription import Subscription
from WMCore.WMBS.Workflow import Workflow
from WMCore.WMSpec.WMWorkload import WMWorkload, WMWorkloadHelper
from WMCore.FwkJobReport.Report import Report
from WMCore.WMExceptions import WM_JOB_ERROR_CODES


def retrieveWMSpec(workflow=None, wmWorkloadURL=None):
    """
    _retrieveWMSpec_

    Given a subscription, this function loads the WMSpec associated with that workload
    """
    if not wmWorkloadURL and workflow:
        wmWorkloadURL = workflow.spec

    if not wmWorkloadURL or not os.path.isfile(wmWorkloadURL):
        logging.error("WMWorkloadURL %s is empty", wmWorkloadURL)
        return None

    wmWorkload = WMWorkloadHelper(WMWorkload("workload"))
    wmWorkload.load(wmWorkloadURL)

    return wmWorkload


def retrieveJobSplitParams(wmWorkload, task):
    """
    _retrieveJobSplitParams_

    Retrieve job splitting parameters from the workflow.  The way this is
    setup currently sucks, we have to know all the job splitting parameters
    up front.  The following are currently supported:
        files_per_job
        min_merge_size
        max_merge_size
        max_merge_events
    """

    # This function has to find the WMSpec, and get the parameters from the spec
    # I don't know where the spec is, but I'll have to find it.
    # I don't want to save it in each workflow area, but I may have to

    if not wmWorkload:
        logging.error("Could not find wmWorkload for splitting")
        return {"files_per_job": 5}
    task = wmWorkload.getTaskByPath(task)
    if not task:
        return {"files_per_job": 5}
    else:
        return task.jobSplittingParameters()


def runSplitter(jobFactory, splitParams):
    """
    _runSplitter_

    Run the jobSplitting as a coroutine method, yielding values as required
    """

    groups = ['test']
    while groups != []:
        groups = jobFactory(**splitParams)
        yield groups
        # Dump it after one go if we're not grabbing by proxy
        if jobFactory.grabByProxy is False:
            break


def capResourceEstimates(jobGroups, constraints):
    """
    _capResourceEstimates_

    Checks the current job resource estimates and cap
    them based on the limits defined in the agent
    config file.
    """
    for jobGroup in jobGroups:
        for j in jobGroup.jobs:
            if not j['estimatedJobTime'] or j['estimatedJobTime'] < constraints['MinWallTimeSecs']:
                j['estimatedJobTime'] = constraints['MinWallTimeSecs']
            if not j['estimatedDiskUsage'] or j['estimatedDiskUsage'] < constraints['MinRequestDiskKB']:
                j['estimatedDiskUsage'] = constraints['MinRequestDiskKB']

            j['estimatedJobTime'] = min(j['estimatedJobTime'], constraints['MaxWallTimeSecs'])
            j['estimatedDiskUsage'] = min(j['estimatedDiskUsage'], constraints['MaxRequestDiskKB'])

            # finally, quantize those
            j['estimatedJobTime'] = quantize(j['estimatedJobTime'], constraints['MinWallTimeSecs'])
            j['estimatedDiskUsage'] = quantize(j['estimatedDiskUsage'], constraints['MinRequestDiskKB'])

    return


def saveJob(job, thisJobNumber, **kwargs):
    """
    _saveJob_

    Actually do the mechanics of saving the job to a pickle file
    """
    job['counter'] = thisJobNumber
    job['spec'] = kwargs.get('workflow').spec
    job['task'] = kwargs.get('wmTaskName')
    job['sandbox'] = kwargs.get('sandbox')
    job['agentNumber'] = kwargs['agentNumber']
    job['agentName'] = kwargs['agentName']
    cacheDir = job.getCache()
    job['cache_dir'] = cacheDir
    job['owner'] = kwargs['owner']
    job['ownerDN'] = kwargs['ownerDN']
    job['ownerGroup'] = kwargs['ownerGroup']
    job['ownerRole'] = kwargs['ownerRole']
    job['scramArch'] = kwargs['scramArch']
    job['swVersion'] = kwargs['swVersion']
    job['numberOfCores'] = kwargs['numberOfCores']
    job['inputDataset'] = kwargs['inputDataset']
    job['inputDatasetLocations'] = kwargs['inputDatasetLocations']
    job['inputPileup'] = kwargs['inputPileup']
    job['allowOpportunistic'] = kwargs['allowOpportunistic']
    job['requiresGPU'] = kwargs['requiresGPU']
    job['gpuRequirements'] = kwargs['gpuRequirements']
    job['requestType'] = kwargs['requestType']
    job['physicsTaskType'] = kwargs['physicsTaskType']
    job['campaignName'] = kwargs['campaignName']

    with open(os.path.join(cacheDir, 'job.pkl'), 'wb') as output:
        pickle.dump(job, output, HIGHEST_PICKLE_PROTOCOL)

    return


def creatorProcess(work, jobCacheDir):
    """
    _creatorProcess_

    Creator work areas and pickle job objects
    """
    createWorkArea = CreateWorkArea()

    try:
        wmbsJobGroup = work.get('jobGroup')
        workflow = work.get('workflow')
        wmWorkload = work.get('wmWorkload')
        work['ownerDN'] = work.get('owner') if work.get('ownerDN', None) is None else work.get('ownerDN')
    except KeyError as ex:
        msg = "Could not find critical key-value in work input.\n"
        msg += str(ex)
        logging.error(msg)
        raise JobCreatorException(msg)
    except Exception as ex:
        msg = "Exception in opening work package. Error: %s" % str(ex)
        logging.exception(msg)
        raise JobCreatorException(msg)

    try:
        createWorkArea.processJobs(jobGroup=wmbsJobGroup,
                                   startDir=jobCacheDir,
                                   workflow=workflow,
                                   wmWorkload=wmWorkload,
                                   cache=False)

        thisJobNumber = work.get('jobNumber', 0)
        for job in wmbsJobGroup.jobs:
            thisJobNumber += 1
            saveJob(job, thisJobNumber, **work)
    except Exception as ex:
        msg = "Exception in processing wmbsJobGroup %i\n. Error: %s" % (wmbsJobGroup.id, str(ex))
        logging.exception(msg)
        raise JobCreatorException(msg)

    return wmbsJobGroup


# This is the code for the multiprocessing based creator
# It's kept around so I can remember how I arranged the exception tree
# Keep this until we make a decision about large-scale transactions
# in the JobCreator.

# def creatorProcess(input, result, jobCacheDir):
#    """
#    _creatorProcess_
#
#    Run the CreateWorkArea code
#    """
#
#
#    createWorkArea  = CreateWorkArea()
#
#
#    while True:
#
#        try:
#            work = input.get()
#        except (EOFError, IOError):
#            crashMessage =  "Hit EOF/IO in getting new work\n"
#            crashMessage += "Assuming this is a graceful break attempt.\n"
#            logging.error(crashMessage)
#            break
#
#        if work == 'STOP':
#            # Put the brakes on
#            break
#
#        try:
#            wmbsJobGroup = work.get('jobGroup')
#            workflow     = work.get('workflow')
#            wmWorkload   = work.get('wmWorkload')
#            wmTaskName   = work.get('wmTaskName')
#            sandbox      = work.get('sandbox')
#            jobNumber    = work.get('jobNumber', 0)
#            wmTaskPrio   = work.get('wmTaskPrio', None)
#        except KeyError, ex:
#            msg =  "Could not find critical key-value in work input.\n"
#            msg += str(ex)
#            logging.error(msg)
#            result.put({status: 'Error', 'msg': msg})
#            continue
#        except Exception, ex:
#            msg =  "Exception in opening work package.\n"
#            msg += str(ex)
#            msg += str(traceback.format_exc())
#            logging.error(msg)
#            result.put({'success': False, 'msg': msg})
#            continue
#
#
#        try:
#            #print "Retrieved work in creatorProcess %i: %s" % (os.getpid(), [x['id'] for x in wmbsJobGroup.jobs])
#            createWorkArea.processJobs(jobGroup = wmbsJobGroup,
#                                       startDir = jobCacheDir,
#                                       workflow = workflow,
#                                       wmWorkload = wmWorkload,
#                                       cache = False)
#
#            for job in wmbsJobGroup.jobs:
#                jobNumber += 1
#                saveJob(job = job, workflow = workflow,
#                        wmTask = wmTaskName,
#                        jobNumber = jobNumber,
#                        wmTaskPrio = wmTaskPrio,
#                        sandbox = sandbox)
#
#            # If we got this far, it should be a success
#            result.put({'success': True, 'jobGroup': wmbsJobGroup})
#
#        except Exception, ex:
#            # Register as failure; move on
#            msg =  "Exception in processing wmbsJobGroup %i\n" % wmbsJobGroup.id
#            msg += str(ex)
#            msg += str(traceback.format_exc())
#            logging.error(msg)
#            result.put({'success': False, 'msg': msg})
#
#        #print "Finished work in creatorProcess %i" % os.getpid()
#        #print [x['id'] for x in wmbsJobGroup.jobs]
#
#        # END while loop that does the work
#
#    return 0









class JobCreatorException(WMException):
    """
    _JobCreatorException_

    Specific JobCreatorPoller exception handling.
    If we ever need it.
    """


class JobCreatorPoller(BaseWorkerThread):
    """
    Poller that does the work of job creation.
    Polls active subscriptions, asks for more work, and checks with local sites.

    """

    def __init__(self, config):
        """
        init jobCreator
        """

        BaseWorkerThread.__init__(self)

        myThread = threading.currentThread()

        # DAO factory for WMBS objects
        self.daoFactory = DAOFactory(package="WMCore.WMBS",
                                     logger=logging,
                                     dbinterface=myThread.dbi)

        self.setBulkCache = self.daoFactory(classname="Jobs.SetCache")
        self.countJobs = self.daoFactory(classname="Jobs.GetNumberOfJobsPerWorkflow")
        self.subscriptionList = self.daoFactory(classname="Subscriptions.ListIncomplete")
        self.setFWJRPath = self.daoFactory(classname="Jobs.SetFWJRPath")

        # information
        self.config = config

        # Variables
        self.defaultJobType = config.JobCreator.defaultJobType
        self.limit = getattr(config.JobCreator, 'fileLoadLimit', 500)
        self.agentNumber = int(getattr(config.Agent, 'agentNumber', 0))
        self.agentName = getattr(config.Agent, 'hostName', '')
        self.glideinLimits = getattr(config.JobCreator, 'GlideInRestriction', None)

        try:
            self.jobCacheDir = getattr(config.JobCreator, 'jobCacheDir',
                                       os.path.join(config.JobCreator.componentDir, 'jobCacheDir'))
            self.check()
        except WMException:
            raise
        except Exception as ex:
            msg = "Unhandled exception while setting up jobCacheDir!\n"
            msg += str(ex)
            logging.error(msg)
            raise JobCreatorException(msg)

        self.changeState = ChangeState(self.config)

        return

    def check(self):
        """
        Initial sanity checks on necessary environment factors

        """

        if not os.path.isdir(self.jobCacheDir):
            if not os.path.exists(self.jobCacheDir):
                os.makedirs(self.jobCacheDir)
            else:
                msg = "Assigned a pre-existant cache object %s.  Failing!" \
                      % (self.jobCacheDir)
                raise JobCreatorException(msg)

    @timeFunction
    def algorithm(self, parameters=None):
        """
        Actually runs the code
        """
        logging.debug("Running JSM.JobCreator")
        try:
            self.pollSubscriptions()
        except WMException:
            # self.close()
            myThread = threading.currentThread()
            if getattr(myThread, 'transaction', False) \
                    and getattr(myThread.transaction, 'transaction', False):
                myThread.transaction.rollback()
            raise
        except Exception as ex:
            # self.close()
            myThread = threading.currentThread()
            if getattr(myThread, 'transaction', False) \
                    and getattr(myThread.transaction, 'transaction', False):
                myThread.transaction.rollback()
            # Handle temporary connection problems (Temporary)
            if "(InterfaceError) not connected" in str(ex):
                logging.error(
                    'There was a connection problem during the JobCreator algorithm, I will try again next cycle')
            else:
                msg = "Failed to execute JobCreator. Error: %s" % str(ex)
                logging.exception(msg)
                raise JobCreatorException(msg)

    def terminate(self, params):
        """
        _terminate_

        Kill the code after one final pass when called by the master thread.
        """
        logging.debug("terminating. doing one more pass before we die")
        self.algorithm(params)

    def pollSubscriptions(self):
        """
        Poller for looking in all active subscriptions for jobs that need to be made.

        """
        logging.info("Beginning JobCreator.pollSubscriptions() cycle.")
        myThread = threading.currentThread()

        # First, get list of Subscriptions
        subscriptions = self.subscriptionList.execute()

        # Okay, now we have a list of subscriptions
        for subscriptionID in subscriptions:
            wmbsSubscription = Subscription(id=subscriptionID)
            try:
                wmbsSubscription.load()
            except IndexError:
                # This happens when the subscription no longer exists
                # i.e., someone executed a kill() function on the database
                # while the JobCreator was in cycle
                # Ignore this subscription
                msg = "JobCreator cannot load subscription %i" % subscriptionID
                logging.error(msg)
                continue

            workflow = Workflow(id=wmbsSubscription["workflow"].id)
            workflow.load()
            wmbsSubscription['workflow'] = workflow
            wmWorkload = retrieveWMSpec(workflow=workflow)

            if not workflow.task or not wmWorkload:
                # Then we have a problem
                # We NEED a sandbox
                # Abort this subscription!
                # But do NOT fail
                # We have no way of marking a subscription as bad per se
                # We'll have to just keep skipping it
                msg = "Have no task for workflow %i\n" % (workflow.id)
                msg += "Aborting Subscription %i" % (subscriptionID)
                logging.error(msg)
                continue

            logging.debug("Have loaded subscription %i with workflow %i\n", subscriptionID, workflow.id)

            # retrieve information from the workload to propagate down to the job configuration
            allowOpport = wmWorkload.getAllowOpportunistic()

            # Set task object
            wmTask = wmWorkload.getTaskByPath(workflow.task)

            # Get generators
            # If you fail to load the generators, pass on the job
            try:
                if hasattr(wmTask.data, 'generators'):
                    manager = GeneratorManager(wmTask)
                    seederList = manager.getGeneratorList()
                else:
                    seederList = []
            except Exception as ex:
                msg = "Had failure loading generators for subscription %i\n" % (subscriptionID)
                msg += "Exception: %s\n" % str(ex)
                msg += "Passing over this error.  It will reoccur next iteration!\n"
                msg += "Please check or remove this subscription!\n"
                logging.error(msg)
                continue

            logging.debug("Going to call wmbsJobFactory for sub %i with limit %i", subscriptionID, self.limit)

            splitParams = retrieveJobSplitParams(wmWorkload, workflow.task)
            logging.debug("Split Params: %s", splitParams)

            # Load the proper job splitting module
            splitterFactory = SplitterFactory(splitParams.get('algo_package', "WMCore.JobSplitting"))
            # and return an instance of the splitting algorithm
            wmbsJobFactory = splitterFactory(package="WMCore.WMBS",
                                             subscription=wmbsSubscription,
                                             generators=seederList,
                                             limit=self.limit)

            # Turn on the jobFactory --> get available files for that subscription, keep result proxies
            wmbsJobFactory.open()

            # Create a function to hold it, calling __call__ from the JobFactory
            # which then calls algorithm method of the job splitting algo instance
            jobSplittingFunction = runSplitter(jobFactory=wmbsJobFactory,
                                               splitParams=splitParams)

            # Now we get to find out how many jobs there are.
            jobNumber = self.countJobs.execute(workflow=workflow.id,
                                               conn=myThread.transaction.conn,
                                               transaction=True)
            jobNumber += splitParams.get('initial_lfn_counter', 0)
            logging.debug("Have %i jobs for workflow %s already in database.", jobNumber, workflow.name)

            while True:
                # This loop runs over the jobFactory,
                # using yield statements and a pre-existing proxy to
                # generate and process new jobs

                # First we need the jobs.
                myThread.transaction.begin()
                try:
                    wmbsJobGroups = next(jobSplittingFunction)
                    logging.info("Retrieved %i jobGroups from jobSplitter", len(wmbsJobGroups))
                except StopIteration:
                    # If you receive a stopIteration, we're done
                    logging.info("Completed iteration over subscription %i", subscriptionID)
                    myThread.transaction.commit()
                    break

                # If we have no jobGroups, we're done
                if len(wmbsJobGroups) == 0:
                    logging.info("Found end in iteration over subscription %i", subscriptionID)
                    myThread.transaction.commit()
                    break

                # Assemble a dict of all the info
                processDict = {'workflow': workflow,
                               'wmWorkload': wmWorkload,
                               'wmTaskName': wmTask.getPathName(),
                               'requestType': wmWorkload.getRequestType(),
                               'jobNumber': jobNumber,
                               'sandbox': wmTask.data.input.sandbox,
                               'owner': wmWorkload.getOwner().get('name', None),
                               'ownerDN': wmWorkload.getOwner().get('dn', None),
                               'ownerGroup': wmWorkload.getOwner().get('vogroup', ''),
                               'ownerRole': wmWorkload.getOwner().get('vorole', ''),
                               'numberOfCores': wmTask.getNumberOfCores(),
                               'requiresGPU': wmTask.getRequiresGPU(),
                               'gpuRequirements': wmTask.getGPURequirements(),
                               'inputDataset': wmTask.getInputDatasetPath(),
                               'inputPileup': wmTask.getInputPileupDatasets(),
                               'swVersion': wmTask.getSwVersion(allSteps=True),
                               'scramArch': wmTask.getScramArch(),
                               'agentNumber': self.agentNumber,
                               'agentName': self.agentName,
                               'allowOpportunistic': allowOpport,
                               'campaignName': wmTask.getCampaignName(),
                               'physicsTaskType': wmTask.getPhysicsTaskType()}

                tempSubscription = Subscription(id=wmbsSubscription['id'])

                # if we have glideinWMS constraints, then adapt all jobs
                if self.glideinLimits:
                    capResourceEstimates(wmbsJobGroups, self.glideinLimits)

                nameDictList = []
                for wmbsJobGroup in wmbsJobGroups:
                    # For each jobGroup, put a dictionary
                    # together and run it with creatorProcess
                    jobsInGroup = len(wmbsJobGroup.jobs)
                    wmbsJobGroup.subscription = tempSubscription
                    tempDict = {}
                    tempDict.update(processDict)
                    tempDict['jobGroup'] = wmbsJobGroup
                    tempDict['jobNumber'] = jobNumber
                    tempDict['inputDatasetLocations'] = wmbsJobGroup.getLocationsForJobs()

                    jobGroup = creatorProcess(work=tempDict,
                                              jobCacheDir=self.jobCacheDir)
                    jobNumber += jobsInGroup

                    # Set jobCache for group
                    for job in jobGroup.jobs:
                        nameDictList.append({'jobid': job['id'],
                                             'cacheDir': job['cache_dir']})
                        job["user"] = wmWorkload.getOwner()["name"]
                        job["group"] = wmWorkload.getOwner()["group"]
                # Set the caches in the database
                try:
                    if len(nameDictList) > 0:
                        self.setBulkCache.execute(jobDictList=nameDictList,
                                                  conn=myThread.transaction.conn,
                                                  transaction=True)
                except WMException:
                    raise
                except Exception as ex:
                    msg = "Unknown exception while setting the bulk cache:\n"
                    msg += str(ex)
                    logging.error(msg)
                    logging.debug("Error while setting bulkCache with following values: %s\n", nameDictList)
                    raise JobCreatorException(msg)

                # Advance the jobGroup in changeState
                for wmbsJobGroup in wmbsJobGroups:
                    self.advanceJobGroup(wmbsJobGroup=wmbsJobGroup)

                # Now end the transaction so that everything is wrapped
                # in a single rollback
                myThread.transaction.commit()

            # END: While loop over jobFactory

            # Close the jobFactory
            wmbsJobFactory.close()

        return

    # This is the code for the multiprocessing based queue retrieval system
    # I'm keeping this here because I hope to go back and re-instate this once
    # I figure out how to deal with the transaction problems.
    # Keep until we can assure ourselves that what we're doing now presents no problems.

    #        while True:
    #            try:
    #                # Get stuff out of the queue with a ridiculously
    #                # short wait time
    #                startTime = time.time()
    #                entry = self.result.get(timeout = self.wait)
    #
    #                if not entry.get('success', False):
    #                    msg =  "Encountered exception processing jobGroup\n"
    #                    msg += entry.get('msg', '')
    #                    logging.error(msg)
    #                    continue
    #                else:
    #                    wmbsJobGroup = entry.get('jobGroup')
    #            except Queue.Empty:
    #                # This means the queue has no current results
    #                stopTime  = time.time()
    #                logging.error("Found empty queue after %f seconds" % (stopTime - startTime))
    #                break
    #
    #            # Set the cache dir
    #            nameDictList = []
    #            for job in wmbsJobGroup.jobs:
    #                nameDictList.append({'jobid':job['id'], 'cacheDir':job['cache_dir']})
    #
    #
    #            try:
    #                myThread.transaction.begin()
    #                self.setBulkCache.execute(jobDictList = nameDictList,
    #                                          conn = myThread.transaction.conn,
    #                                          transaction = True)
    #                myThread.transaction.commit()
    #            except Exception, ex:
    #                msg =  "Unhandled exception while setting jobCache for jobGroup %i\n" % wmbsJobGroup.id
    #                msg += str(ex)
    #                logging.error(msg)
    #                raise JobCreatorException(msg)
    #
    #            self.advanceJobGroup(wmbsJobGroup = wmbsJobGroup)
    #
    #            logging.debug("Finished call for jobGroup %i" \
    #                          % (wmbsJobGroup.id))
    #
    #        #END: While loop over wmbsJob
    #
    #        return


    def advanceJobGroup(self, wmbsJobGroup):
        """
        _advanceJobGroup_

        Mark jobGroup as ready in changeState
        """
        try:
            self.changeState.propagate(wmbsJobGroup.jobs, 'created', 'new')

            createFailedJobs = [x for x in wmbsJobGroup.jobs if x.get('failedOnCreation', False)]
            self.generateCreateFailedReports(createFailedJobs)
            self.changeState.propagate(createFailedJobs, 'createfailed', 'created')
        except WMException:
            raise
        except Exception as ex:
            msg = "Unhandled exception while calling changeState.\n"
            msg += str(ex)
            logging.error(msg)
            logging.debug("Error while using changeState on the following jobs: %s\n", wmbsJobGroup.jobs)

        logging.info("JobCreator has finished creating jobGroup %i.\n", wmbsJobGroup.id)
        return

    def generateCreateFailedReports(self, createFailedJobs):
        """
        _generateCreateFailedReports_

        Create and store FWJR for the  jobs that failed on creation
        leaving meaningful information about what happened with them
        """
        if not createFailedJobs:
            return

        fjrsToSave = []
        for failedJob in createFailedJobs:
            report = Report()
            report.addError("CreationFailure", failedJob["failedErrCode"], "CreationFailure",
                            failedJob.get("failedReason", WM_JOB_ERROR_CODES[failedJob["failedErrCode"]]))
            jobCache = failedJob.getCache()
            try:
                fjrPath = os.path.join(jobCache, "Report.0.pkl")
                report.save(fjrPath)
                fjrsToSave.append({"jobid": failedJob["id"], "fwjrpath": fjrPath})
                failedJob["fwjr"] = report
            except Exception:
                logging.error("Something went wrong while saving the report for  job %s", failedJob["id"])

        myThread = threading.currentThread()
        self.setFWJRPath.execute(binds=fjrsToSave, conn=myThread.transaction.conn, transaction=True)

        return



#!/usr/bin/env python
"""
_DeleteFile_

Oracle implementation of File.Delete

"""
__all__ = []



from WMCore.WMBS.MySQL.Files.Delete import Delete as DeleteFileMySQL

class Delete(DeleteFileMySQL):
    sql = DeleteFileMySQL.sql



"""
File       : MSPileupObj.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: MSPileupObj module provides MSPileup data structure:

{
    "pileupName": string with the pileup dataset (madatory)
    "pileupType": string with a constant value (mandatory)
    "insertTime": int, seconds since epoch in GMT timezone (service-based)
    "lastUpdateTime": int, seconds since epoch in GMT timezone (service-based)
    "expectedRSEs": ["Disk1", "Disk2", etc], a non-empty list of strings with the RSE names (mandatory)
    "currentRSEs": ["Disk1", "Disk3"], a list of strings with the RSE names (service-based)
    "fullReplicas": integer,  # total number of replicas to keep on Disk (optional)
    "campaigns": [ "name", ... ] # list of workflow campaigns using this pileup (optional)
    "containerFraction": real number with the container fraction to be distributed (optional)
    "replicationGrouping": string with a constant value (DATASET or ALL, (optional)
    "activatedOn": int, seconds since epoch in GMT timezone (service-based)
    "deactivatedOn": int, seconds since epoch in GMT timezone (service-based)
    "active": boolean, (mandatory)
    "pileupSize": integer, current size of the pileup in bytes (service-based)
    "customName": string, custom container DID (optional)
    "transition": [{'updateTime': 123, 'containerFraction': 0.5, 'customDID': 'blah', 'DN': 'blah2'}, ...]
                  list of transition records for partial data placement
    "ruleIds: list of strings (rules) used to lock the pileup id (service-based)
}

The data flow should be done via list of objects, e.g.
[<pileup object 1>, <pileup object 2>, ..., <pileup object n>]
"""

# system modules
import json

# WMCore modules
from Utils.Timers import gmtimeSeconds
from WMCore.MicroService.Tools.Common import getMSLogger
from WMCore.Lexicon import dataset


class MSPileupObj():
    """
    MSPileupObj defines MSPileup data stucture
    """
    def __init__(self, pdict, verbose=None, logger=None, validRSEs=None):
        self.logger = getMSLogger(verbose, logger)
        if not validRSEs:
            validRSEs = []
        self.validRSEs = validRSEs

        self.data = {
            'pileupName': pdict['pileupName'],
            'pileupType': pdict['pileupType'],
            'insertTime': pdict.get('insertTime', gmtimeSeconds()),
            'lastUpdateTime': pdict.get('lastUpdateTime', gmtimeSeconds()),
            'expectedRSEs': pdict['expectedRSEs'],
            'currentRSEs': pdict.get('currentRSEs', []),
            'fullReplicas': pdict.get('fullReplicas', 1),
            'campaigns': pdict.get('campaigns', []),
            'containerFraction': pdict.get('containerFraction', 1.0),
            'replicationGrouping': pdict.get('replicationGrouping', 'ALL'),
            'activatedOn': pdict.get('activatedOn', gmtimeSeconds()),
            'deactivatedOn': pdict.get('deactivatedOn', gmtimeSeconds()),
            'active': pdict['active'],
            'pileupSize': pdict.get('pileupSize', 0),
            'customName': pdict.get('customName', ''),
            'transition': pdict.get('transition', []),
            'ruleIds': pdict.get('ruleIds', [])}
        valid, msg = self.validate(self.data)
        if not valid:
            msg = f'MSPileup input is invalid, {msg}'
            raise Exception(msg)

    def __str__(self):
        """
        Return human readable representation of pileup data
        """
        return json.dumps(self.data, indent=4)

    def getPileupData(self):
        """
        Get pileup data
        """
        return self.data

    def validate(self, pdict=None):
        """
        Validate data according to its schema. If data is not provided via
        input pdict parameter, the validate method will validate internal
        data object.

        :param pdict: input data dictionary (optional)
        :return: (boolean status, string message) result of validation
        """
        msg = ""
        if not pdict:
            pdict = self.data
        docSchema = schema()
        if set(pdict) != set(docSchema):
            pkeys = set(pdict.keys())
            skeys = set(docSchema.keys())
            msg = f"provided object {pkeys} keys are not equal to schema keys {skeys}"
            self.logger.error(msg)
            return False, msg
        for key, val in pdict.items():
            if key not in docSchema:
                msg = f"Failed to validate {key}, not found in {docSchema}"
                self.logger.error(msg)
                return False, msg
            _, stype = docSchema[key]  # expected data type for our key
            if not isinstance(val, stype):
                dtype = str(type(val))     # obtained data type of our value
                msg = f"Failed to validate: {key}, expect data-type {stype} got type {dtype}"
                self.logger.error(msg)
                return False, msg
            if key == 'pileupName' or (key == 'customName' and val != ''):
                if key == 'customName':
                    # we may have additional suffix which we should strip off
                    # before validation, see customDID definition in
                    # WMCore/MicroService/MSPileup/MSPileupData.py
                    val = val.split('-V')[:-1][0]
                try:
                    dataset(val)
                except AssertionError:
                    msg = f"{key} value {val} does not match dataset pattern"
                    self.logger.error(msg)
                    return False, msg
            if key == "pileupType" and val not in ['classic', 'premix']:
                msg = f"pileupType value {val} is neither of ['classic', 'premix']"
                self.logger.error(msg)
                return False, msg
            if key == 'replicationGrouping' and val not in ['DATASET', 'ALL']:
                msg = f"replicationGrouping value {val} is neither of ['DATASET', 'ALL']"
                self.logger.error(msg)
                return False, msg
            if key == 'containerFraction' and (val > 1 or val <= 0):
                msg = f"containerFraction value {val} outside (0,1] range"
                self.logger.error(msg)
                return False, msg
            if key == 'transition' and not self.validateTransitionRecord(val, pdict):
                msg = f"transition record {val} is invalid"
                self.logger.error(msg)
                return False, msg
            if key in ('expectedRSEs', 'currentRSEs') and not self.validateRSEs(val):
                msg = f"{key} value {val} is not in validRSEs {self.validRSEs}"
                self.logger.error(msg)
                return False, msg
            if key == 'expectedRSEs' and len(val) == 0:
                msg = 'document require non-empty list of expectedRSEs'
                self.logger.error(msg)
                return False, msg
        return True, msg

    def validateTransitionRecord(self, records, pdoc):
        """
        Validate given transition records

        :param records: transition records
        :param pdoct: pileup document
        :return: boolean
        """
        for record in records:
            keys = ['updateTime', 'customDID', 'DN', 'containerFraction']
            for key in keys:
                if key not in record:
                    self.logger.error("key '%s' is not present in transition record %s", key, record)
                    return False
                val = record[key]
                if key == 'updateTime' and (val < pdoc['insertTime'] or val > gmtimeSeconds()):
                    # the update time should be in range between pileup insert time and now
                    self.logger.error("wrong value '%s' for updateTime in transition record %s", val, record)
                    return False
                if key in ['customDID', 'DN'] and not isinstance(val, str):
                    self.logger.error("wrong data-type of value '%s' for {key} in transition record %s", val, record)
                    return False
                if key == 'containerFraction' and (val > 1 or val <= 0):
                    self.logger.error("wrong value '%s' for containerFraction in transition record %s", val, record)
                    return False
        return True

    def validateRSEs(self, rseList):
        """
        Validate given list of RSEs

        :param rseList: list of RSEs
        :return: boolean
        """
        if rseList == self.validRSEs:
            return True
        for rse in rseList:
            if rse not in self.validRSEs:
                return False
        return True


def schema():
    """
    Return the data schema for a record in MongoDB.
    It's a dictionary where:
    - key is schema attribute name
    - a value is a tuple of (default value, expected data type)

    :return: a dictionary
    """
    doc = {'pileupName': ('', str),
           'pileupType': ('', str),
           'insertTime': (0, int),
           'lastUpdateTime': (0, int),
           'expectedRSEs': ([], list),
           'currentRSEs': ([], list),
           'fullReplicas': (0, int),
           'campaigns': ([], list),
           'containerFraction': (1.0, float),
           'replicationGrouping': ('', str),
           'activatedOn': (0, int),
           'deactivatedOn': (0, int),
           'active': (False, bool),
           'pileupSize': (0, int),
           'customName': ('', str),
           'transition': ([], list),
           'ruleIds': ([], list)}
    return doc



#!/usr/bin/env python
# encoding: utf-8
"""
Decorators_t.py

Created by Dave Evans on 2010-07-29.
Copyright (c) 2010 Fermilab. All rights reserved.
"""

from builtins import object

import unittest
import WMCore.GroupUser.Decorators as Decs

class CrashTestDummy(object):

    def __init__(self):
        self._connected = False
        self.group = None
        self.owner = None

    connected = property(lambda x: x._connected)

    def connect(self):
        self._connected = True

    @Decs.requireConnection
    def __call__(self, x):
        return x
    @Decs.requireGroup
    def getGroup(self):
        return self.group

    @Decs.requireUser
    def getOwner(self):
        return self.owner

class Decorators_t(unittest.TestCase):

    def testA(self):
        """test requireConnection"""

        testDummy = CrashTestDummy()
        self.assertFalse(testDummy.connected)
        testDummy(99)
        self.assertTrue(testDummy.connected)

    def testB(self):
        """test requireGroup"""

        testDummy = CrashTestDummy()

        self.assertRaises(Exception, testDummy.getGroup)

        testDummy.group = "DMWM"

        self.assertEqual(testDummy.getGroup(), "DMWM")

    def testC(self):
        """test requireUser"""

        testDummy = CrashTestDummy()

        self.assertRaises(Exception, testDummy.getOwner)

        testDummy.owner = "TheRookie"

        self.assertEqual(testDummy.getOwner(), "TheRookie")


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
testing file manager

"""
from __future__ import print_function

import logging
import os.path
import shutil
import tempfile
import unittest

import WMCore.Storage.StageOutError
from WMCore.Storage.FileManager import StageInMgr, StageOutMgr, DeleteMgr


class FileManagerTest(unittest.TestCase):
    def setUp(self):
        self.testDir = None

    def tearDown(self):
        if (self.testDir != None):
            try:
                shutil.rmtree(self.testDir)
            except Exception:
                # meh, if it fails, I guess something weird happened
                pass

    def testStageFile(self):
        pass

    # def stageFile(self, fileToStage, stageOut = True):
    def testDelete(self):
        pass

    # def deleteLFN(self, lfn):
    def testInitialiseSiteConf(self):
        pass

    # def initialiseSiteConf(self):
    def testInitialiseOverride(self):
        # def initialiseOverride(self):
        pass

    def testGetTransferDetails(self):
        pass

    # def getTransferDetails(self, lfn, currentMethod):
    def testStageIn(self):
        pass

    def testStageOut(self):
        pass

    # def stageIn(self,fileToStage):
    # def stageOut(self,fileToStage):

    def test_doTransfer(self):
        pass

    # def _doTransfer(self, currentMethod, methodCounter, lfn, pfn, stageOut):



    def testCleanSuccessfulStageOuts(self):
        pass

    # def cleanSuccessfulStageOuts(self):
    def testSearchTFC(self):
        pass

    # def searchTFC(self, lfn):

    def testStageOutMgrWrapperWin(self):
        fileForTransfer = {'LFN': '/etc/hosts', \
                           'PFN': 'file:///etc/hosts', \
                           'PNN': None, \
                           'StageOutCommand': None}
        wrapper = StageOutMgr(**{
            'command': 'test-win',
            'option': '',
            'phedex-node': 'test-win',
            'lfn-prefix': ''})
        wrapper(fileForTransfer)

    def testStageOutMgrWrapperFail(self):
        fileForTransfer = {'LFN': 'failtest', \
                           'PFN': 'failtest', \
                           'PNN': None, \
                           'StageOutCommand': None}
        wrapper = StageOutMgr(numberOfRetries=1,
                              retryPauseTime=0, **{
                'command': 'test-fail',
                'option': '',
                'phedex-node': 'test-win',
                'lfn-prefix': ''})
        self.assertRaises(WMCore.Storage.StageOutError.StageOutError, wrapper.__call__, fileForTransfer)

    def testStageOutMgrWrapperRealCopy(self):
        self.testDir = tempfile.mkdtemp()
        fileForTransfer = {'LFN': '/etc/hosts', \
                           'PFN': '/etc/hosts', \
                           'PNN': None, \
                           'StageOutCommand': None}
        wrapper = StageOutMgr(**{
            'command': 'cp',
            'option': '',
            'phedex-node': 'test-win',
            'lfn-prefix': self.testDir})
        wrapper(fileForTransfer)
        self.assertTrue(os.path.exists(os.path.join(self.testDir, '/etc/hosts')))

    def testStageOutMgrWrapperRealCopyFallback(self):
        self.testDir = tempfile.mkdtemp()
        fileForTransfer = {'LFN': '/etc/hosts', \
                           'PFN': '/etc/hosts', \
                           'PNN': None, \
                           'StageOutCommand': None}
        wrapper = StageOutMgr(**{
            'command': 'testFallbackToOldBackend',
            'option': '',
            'phedex-node': 'test-win',
            'lfn-prefix': self.testDir})
        wrapper(fileForTransfer)
        self.assertTrue(os.path.exists(os.path.join(self.testDir, '/etc/hosts')))

    def testStageInMgrWrapperWin(self):
        fileForTransfer = {'LFN': '/etc/hosts', \
                           'PFN': '/etc/hosts', \
                           'PNN': None, \
                           'StageOutCommand': None}
        wrapper = StageInMgr(**{
            'command': 'test-win',
            'option': '',
            'phedex-node': 'test-win',
            'lfn-prefix': ''})
        wrapper(fileForTransfer)

    def testStageInMgrWrapperFail(self):
        fileForTransfer = {'LFN': 'failtest', \
                           'PFN': 'failtest', \
                           'PNN': None, \
                           'StageOutCommand': None}
        wrapper = StageInMgr(numberOfRetries=1,
                             retryPauseTime=0, **{
                'command': 'test-fail',
                'option': '',
                'phedex-node': 'test-win',
                'lfn-prefix': ''})
        self.assertRaises(WMCore.Storage.StageOutError.StageOutError, wrapper.__call__, fileForTransfer)

    def testStageInMgrWrapperRealCopy(self):

        self.testDir = tempfile.mkdtemp()
        shutil.copy('/etc/hosts', self.testDir + '/INPUT')
        fileForTransfer = {'LFN': '/INPUT', \
                           'PFN': '%s/etc/hosts' % self.testDir, \
                           'PNN': None, \
                           'StageOutCommand': None}

        wrapper = StageInMgr(**{
            'command': 'cp',
            'option': '',
            'phedex-node': 'test-win',
            'lfn-prefix': self.testDir})
        wrapper(fileForTransfer)

    def testStageInMgrWrapperRealCopyFallback(self):
        self.testDir = tempfile.mkdtemp()
        shutil.copy('/etc/hosts', self.testDir + '/INPUT')
        fileForTransfer = {'LFN': '/INPUT', \
                           'PFN': '%s/etc/hosts' % self.testDir, \
                           'PNN': None, \
                           'StageOutCommand': None}

        wrapper = StageInMgr(**{
            'command': 'testFallbackToOldBackend',
            'option': '',
            'phedex-node': 'test-win',
            'lfn-prefix': self.testDir})
        wrapper(fileForTransfer)

    def testDeleteMgrWrapper(self):
        self.testDir = tempfile.mkdtemp()
        shutil.copy('/etc/hosts', self.testDir + '/INPUT')
        fileForTransfer = {'LFN': '/INPUT', \
                           'PFN': '%s/etc/hosts' % self.testDir, \
                           'PNN': None, \
                           'StageOutCommand': None}

        wrapper = StageInMgr(**{
            'command': 'cp',
            'option': '',
            'phedex-node': 'test-win',
            'lfn-prefix': self.testDir})
        retval = wrapper(fileForTransfer)
        print("got the retval %s" % retval)
        wrapper = DeleteMgr(**{
            'command': 'cp',
            'option': '',
            'phedex-node': 'test-win',
            'lfn-prefix': self.testDir})
        wrapper(retval)


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    unittest.main()



from __future__ import (print_function, division)

import argparse

import sys


def main(opts):
    if opts.text:
        print(opts.text)
    if opts.exception:
        print(opts.exception)
        raise Exception(opts.exception)
    if opts.exit:
        sys.exit(opts.exit)


if __name__=="__main__":
    if __name__ == "__main__":
        parser = argparse.ArgumentParser()
        parser.add_argument("-text", dest="text", action='store')
        parser.add_argument("-exception", dest="exception", action='store')
        parser.add_argument("-exit", dest="exit", action='store')
        opts = parser.parse_args()
        main(opts)



#!/usr/bin/env python
"""
This unit test isn't really a unit test, more an explaination of how to use the
CMSCouch library to work with CouchDB. It currently assumes you have a running
CouchDB instance, and is not going to work in an automated way just yet - we'll
need to add Couch as an external, include it in start up scripts etc.
"""
import unittest
import os
import hashlib
import base64
import time
from Utils.Utilities import encodeUnicodeToBytes
from WMCore.Database.CMSCouch import (CouchServer, CouchMonitor, Document, Database,
                                      CouchInternalServerError, CouchNotFoundError)


class CMSCouchTest(unittest.TestCase):

    def setUp(self):
        # Make an instance of the server
        creds = os.getenv("COUCHURL", 'http://admin:password@localhost:5984')
        self.server = CouchServer(creds)
        testname = self.id().split('.')[-1]
        # Create a database, drop an existing one first
        self.testdbname = 'cmscouch_unittest_%s' % testname.lower()

        if self.testdbname in self.server.listDatabases():
            self.server.deleteDatabase(self.testdbname)

        self.server.createDatabase(self.testdbname)
        self.db = self.server.connectDatabase(self.testdbname)
        self.replicatordb = self.server.connectDatabase('_replicator')

    def tearDown(self):
        """Destroy our testing setup"""
        if self.testdbname in self.server.listDatabases():
            self.server.deleteDatabase(self.testdbname)

    def testCommitOne(self):
        # Can I commit one dict
        doc = {'foo':123, 'bar':456}
        id = self.db.commitOne(doc)[0]['id']
        # What about a Document
        doc = Document(inputDict = doc)
        id = self.db.commitOne(doc)[0]['id']

    def testCommitOneWithQueue(self):
        """
        CommitOne bypasses the queue, but it should maintain the queue if
        present for a future call to commit.
        """
        # Queue up five docs
        doc = {'foo':123, 'bar':456}
        for i in range(1,6):
            self.db.queue(doc)
        # Commit one Document
        doc = Document(inputDict = doc)
        id = self.db.commitOne(doc)[0]['id']
        self.assertEqual(1, len(self.db.allDocs()['rows']))
        self.db.commit()
        self.assertEqual(6, len(self.db.allDocs()['rows']))

    def testTimeStamping(self):
        doc = {'foo':123, 'bar':456}
        id = self.db.commitOne(doc, timestamp=True)[0]['id']
        doc = self.db.document(id)
        self.assertTrue('timestamp' in doc.keys())

    def testDeleteDoc(self):
        doc = {'foo':123, 'bar':456}
        self.db.commitOne(doc)
        all_docs = self.db.allDocs()
        self.assertEqual(1, len(all_docs['rows']))

        # The db.delete_doc is immediate
        id = all_docs['rows'][0]['id']
        self.db.delete_doc(id)
        all_docs = self.db.allDocs()
        self.assertEqual(0, len(all_docs['rows']))

    def testDeleteQueuedDocs(self):
        doc1 = {'foo':123, 'bar':456}
        doc2 = {'foo':789, 'bar':101112}
        self.db.queue(doc1)
        self.db.queue(doc2)
        self.db.commit()

        all_docs = self.db.allDocs()
        self.assertEqual(2, len(all_docs['rows']))
        for res in all_docs['rows']:
            id = res['id']
            doc = self.db.document(id)
            self.db.queueDelete(doc)
        all_docs = self.db.allDocs()
        self.assertEqual(2, len(all_docs['rows']))

        self.db.commit()

        all_docs = self.db.allDocs()
        self.assertEqual(0, len(all_docs['rows']))

    def testReplicate(self):
        repl_db = self.server.connectDatabase(self.db.name + 'repl')

        doc_id = self.db.commitOne({'foo':123}, timestamp=True)[0]['id']
        doc_v1 = self.db.document(doc_id)

        #replicate
        resp = self.server.replicate(self.db.name, repl_db.name, sleepSecs=5)
        self.assertTrue(resp["ok"])
        self.assertTrue("id" in resp)
        self.assertTrue("rev" in resp)

        # wait for a few seconds to replication to be triggered.
        time.sleep(1)
        self.assertEqual(self.db.document(doc_id), repl_db.document(doc_id))
        self.server.deleteDatabase(repl_db.name)

    def testSlashInDBName(self):
        """
        Slashes are a valid character in a database name, and are useful as it
        creates a directory strucutre for the couch data files.
        """
        db_name = 'wmcore/unittests'
        try:
            self.server.deleteDatabase(db_name)
        except:
            # Ignore this - the database shouldn't already exist
            pass

        db = self.server.createDatabase(db_name)
        info = db.info()
        assert info['db_name'] == db_name

        db_name = 'wmcore/unittests'
        db = self.server.connectDatabase(db_name)
        info = db.info()
        assert info['db_name'] == db_name

        db = Database(db_name, url = os.environ["COUCHURL"])
        info = db.info()
        assert info['db_name'] == db_name

        self.server.deleteDatabase(db_name)

    def testInvalidName(self):
        """
        Capitol letters are not allowed in database names.
        """
        db_name = 'Not A Valid Name'
        self.assertRaises(ValueError, self.server.createDatabase, db_name)
        self.assertRaises(ValueError, self.server.deleteDatabase, db_name)
        self.assertRaises(ValueError, self.server.connectDatabase, db_name)
        self.assertRaises(ValueError, Database, db_name)

    def testDocumentSerialisation(self):
        """
        A document should be writable into the couchdb with a timestamp.
        """
        d = Document()
        d['foo'] = 'bar'
        doc_info = self.db.commit(doc=d, timestamp=True)[0]
        d_from_db = self.db.document(doc_info['id'])
        self.assertEqual(d['foo'], d_from_db['foo'])
        self.assertEqual(d['timestamp'], d_from_db['timestamp'])

    def testAttachments(self):
        """
        Test uploading attachments with and without checksumming
        """
        doc = self.db.commitOne({'foo':'bar'}, timestamp=True)[0]
        attachment1 = "Hello"
        attachment2 = "How are you today?"
        attachment3 = "I'm very well, thanks for asking"
        attachment4 = "Lovely weather we're having"
        attachment5 = "Goodbye"
        keyhash = hashlib.md5()
        keyhash.update(encodeUnicodeToBytes(attachment5))
        attachment5_md5 = keyhash.digest()
        attachment5_md5 = base64.b64encode(attachment5_md5)
        attachment6 = "Good day to you, sir!"
        #TODO: add a binary attachment - e.g. tar.gz
        doc = self.db.addAttachment(doc['id'], doc['rev'], attachment1)
        doc = self.db.addAttachment(doc['id'], doc['rev'], attachment2, contentType="foo/bar")
        doc = self.db.addAttachment(doc['id'], doc['rev'], attachment3, name="my_greeting")
        doc = self.db.addAttachment(doc['id'], doc['rev'], attachment4, add_checksum=True)
        doc = self.db.addAttachment(doc['id'], doc['rev'], attachment5, checksum=attachment5_md5)

        self.assertRaises(CouchInternalServerError, self.db.addAttachment, doc['id'], doc['rev'], attachment6, checksum='123')

    def testRevisionHandling(self):
        # This test won't work from an existing database, conflicts will be preserved, so
        # ruthlessly remove the databases to get a clean slate.
        try:
            self.server.deleteDatabase(self.db.name)
        except CouchNotFoundError:
            pass # Must have been deleted already

        try:
            self.server.deleteDatabase(self.db.name + 'repl')
        except CouchNotFoundError:
            pass # Must have been deleted already

        # I'm going to create a conflict, so need a replica db
        self.db = self.server.connectDatabase(self.db.name)
        repl_db = self.server.connectDatabase(self.db.name + 'repl')

        doc_id = self.db.commitOne({'foo':123}, timestamp=True)[0]['id']
        doc_v1 = self.db.document(doc_id)

        #replicate
        self.server.replicate(self.db.name, repl_db.name, sleepSecs=5)
        time.sleep(1)

        doc_v2 = self.db.document(doc_id)
        doc_v2['bar'] = 456
        doc_id_rev2 = self.db.commitOne(doc_v2)[0]
        doc_v2 = self.db.document(doc_id)

        #now update the replica
        conflict_doc = repl_db.document(doc_id)
        conflict_doc['bar'] = 101112
        repl_db.commitOne(conflict_doc)

        #replicate, creating the conflict
        self.server.replicate(self.db.name, repl_db.name, sleepSecs=5)
        time.sleep(1)

        tempDesignDoc = {'views': {
                             'conflicts': {
                                 'map': "function(doc) {if(doc._conflicts) {emit(doc._conflicts, null);}}"
                                           },
                                   }
                         }
        # create the temporary views, which in CouchDB 3.x must be through a
        # permanent view (within a design doc)
        repl_db.put('/%s/_design/TempDesignDoc' % repl_db.name, tempDesignDoc)
        self.db.put('/%s/_design/TempDesignDoc' % self.db.name, tempDesignDoc)

        # Should have one conflict in the repl database
        dataRepl = repl_db.get('/%s/_design/TempDesignDoc/_view/conflicts' % repl_db.name)
        self.assertEqual(dataRepl['total_rows'], 1)
        # Should have no conflicts in the source database
        data = self.db.get('/%s/_design/TempDesignDoc/_view/conflicts' % self.db.name)
        self.assertEqual(data['total_rows'], 0)
        self.assertTrue(repl_db.documentExists(dataRepl['rows'][0]['id'], rev=dataRepl['rows'][0]['key'][0]))

        repl_db.delete_doc(dataRepl['rows'][0]['id'], rev=dataRepl['rows'][0]['key'][0])
        dataRepl = repl_db.get('/%s/_design/TempDesignDoc/_view/conflicts' % repl_db.name)

        self.assertEqual(dataRepl['total_rows'], 0)
        self.server.deleteDatabase(repl_db.name)

        #update it again
        doc_v3 = self.db.document(doc_id)
        doc_v3['baz'] = 789
        doc_id_rev3 = self.db.commitOne(doc_v3)[0]
        doc_v3 = self.db.document(doc_id)

        #test that I can pull out an old revision
        doc_v1_test = self.db.document(doc_id, rev=doc_v1['_rev'])
        self.assertEqual(doc_v1, doc_v1_test)

        #test that I can check a revision exists
        self.assertTrue(self.db.documentExists(doc_id, rev=doc_v2['_rev']))

        self.assertFalse(self.db.documentExists(doc_id, rev='1'+doc_v2['_rev']))

        #why you shouldn't rely on rev
        self.db.compact(blocking=True)
        self.assertFalse(self.db.documentExists(doc_id, rev=doc_v1['_rev']))
        self.assertFalse(self.db.documentExists(doc_id, rev=doc_v2['_rev']))
        self.assertTrue(self.db.documentExists(doc_id, rev=doc_v3['_rev']))

    def testCommit(self):
        """
        Test queue and commit modes
        """
        # try to commit 2 random docs
        doc = {'foo':123, 'bar':456}
        self.db.queue(doc)
        self.db.queue(doc)
        self.assertEqual(2, len(self.db.commit()))

        # committing 2 docs with the same id will fail
        self.db.queue(Document(id = "1", inputDict = {'foo':123, 'bar':456}))
        self.db.queue(Document(id = "1", inputDict = {'foo':1234, 'bar':456}))
        answer = self.db.commit()
        self.assertEqual(2, len(answer))
        self.assertEqual(answer[0]['ok'], True)
        self.assertEqual(answer[1]['error'], 'conflict')

        # bulk commit with one conflict unresolved
        self.db.queue(Document(id = "2", inputDict = doc))
        self.db.queue(Document(id = "2", inputDict = {'foo':1234, 'bar':456}))
        answer = self.db.commit()
        self.assertEqual(2, len(answer))
        self.assertTrue('error' not in answer[0])
        self.assertTrue('error' in answer[1])
        self.assertEqual(answer[0]['id'], '2')
        self.assertEqual(answer[1]['id'], '2')

        # callbacks can do stuff when conflicts arise
        # this particular one just overwrites the document
        def callback(db, data, result):
            for doc in data['docs']:
                if doc['_id'] == result['id']:
                    doc['_rev'] = db.document(doc['_id'])['_rev']
                    retval = db.commitOne(doc)
            return retval[0]

        # bulk commit with one callback enabled (conflicts resolved)
        self.db.queue(Document(id = "2", inputDict = {'foo':5, 'bar':6}))
        self.db.queue(Document(id = "3", inputDict = doc))
        answer = self.db.commit(callback = callback)
        self.assertEqual(2, len(answer))
        self.assertTrue('error' not in answer[0])
        self.assertTrue('error' not in answer[1])
        self.assertEqual(answer[0]['id'], '2')
        self.assertEqual(answer[1]['id'], '3')
        updatedDoc = self.db.document('2')
        self.assertEqual(updatedDoc['foo'], 5)
        self.assertEqual(updatedDoc['bar'], 6)

        return

    def testUpdateHandler(self):
        """
        Test that update function support works
        """

        update_ddoc = {
            '_id':'_design/foo',
            'language': 'javascript',
            'updates':{
                "bump-counter" : 'function(doc, req) {if (!doc.counter) {doc.counter = 0};doc.counter += 1;return [doc,"bumped it!"];}',
            }
        }
        self.db.commit(update_ddoc)
        doc = {'foo': 123, 'counter': 0}
        doc_id = self.db.commit(doc)[0]['id']
        self.assertEqual("bumped it!", self.db.updateDocument(doc_id, 'foo', 'bump-counter'))

        self.assertEqual(1, self.db.document(doc_id)['counter'])


    def testList(self):
        """
        Test list function works ok
        """
        update_ddoc = {
            '_id':'_design/foo',
            'language': 'javascript',
            'views' : {
                       'all' : {
                                'map' : 'function(doc) {emit(null, null) }'
                                },
                       },
            'lists' : {
                'errorinoutput' : 'function(doc, req) {send("A string with the word error in")}',
                'malformed' : 'function(doc, req) {somethingtoraiseanerror}',
            }
        }
        self.db.commit(update_ddoc)
        # approriate errors raised
        self.assertRaises(CouchNotFoundError, self.db.loadList, 'foo', 'error', 'view_doesnt_exist')
        self.assertRaises(CouchInternalServerError, self.db.loadList, 'foo', 'malformed', 'all')
        # error in list output string shouldn't raise an error
        self.assertEqual(self.db.loadList('foo', 'errorinoutput', 'all'),
                         "A string with the word error in")

    def testAllDocs(self):
        """
        Test AllDocs with options
        """
        self.db.queue(Document(id = "1", inputDict = {'foo':123, 'bar':456}))
        self.db.queue(Document(id = "2", inputDict = {'foo':123, 'bar':456}))
        self.db.queue(Document(id = "3", inputDict = {'foo':123, 'bar':456}))

        self.db.commit()
        self.assertEqual(3, len(self.db.allDocs()['rows']))
        self.assertEqual(2, len(self.db.allDocs({'startkey': "2"})['rows']))
        self.assertEqual(2, len(self.db.allDocs(keys = ["1", "3"])['rows']))
        self.assertEqual(1, len(self.db.allDocs({'limit':1}, ["1", "3"])['rows']))
        self.assertTrue('error' in self.db.allDocs(keys = ["1", "4"])['rows'][1])

    def testUpdateBulkDocuments(self):
        """
        Test AllDocs with options
        """
        self.db.queue(Document(id="1", inputDict={'foo':123, 'bar':456}))
        self.db.queue(Document(id="2", inputDict={'foo':123, 'bar':456}))
        self.db.queue(Document(id="3", inputDict={'foo':123, 'bar':456}))
        self.db.commit()

        self.db.updateBulkDocumentsWithConflictHandle(["1", "2", "3"], {'foo': 333}, 2)
        result = self.db.allDocs({"include_docs": True})['rows']
        self.assertEqual(3, len(result))
        for item in result:
            self.assertEqual(333, item['doc']['foo'])

        self.db.updateBulkDocumentsWithConflictHandle(["1", "2", "3"], {'foo': 222}, 10)
        result = self.db.allDocs({"include_docs": True})['rows']
        self.assertEqual(3, len(result))
        for item in result:
            self.assertEqual(222, item['doc']['foo'])

    def testUpdateHandlerAndBulkUpdateProfile(self):
        """
        Test that update function support works
        """
        # for actual test increase the size value: For 10000 records, 96 sec vs 4 sec
        size = 100
        for i in range(size):
            self.db.queue(Document(id="%s" % i, inputDict={'name':123, 'counter':0}))

        update_doc = {
            '_id':'_design/foo',
            'language': 'javascript',
            'updates':{
                "change-counter" : """function(doc, req) { if (doc) { var data = JSON.parse(req.body);
                                      for (var field in data) {doc.field = data.field;} return [doc, 'OK'];}}""",
            }
        }

        self.db.commit(update_doc)
        start = time.time()
        for id in range(size):
            doc_id = "%s" % id
            self.db.updateDocument(doc_id, 'foo', 'change-counter', {'counter': 1}, useBody=True)
        end = time.time()

        print("update handler: %s sec" % (end - start))

        start = time.time()
        ids = []
        for id in range(size):
            doc_id = "%s" % id
            ids.append(doc_id)
        self.db.updateBulkDocumentsWithConflictHandle(ids, {'counter': 2}, 1000)
        end = time.time()

        print("bulk update: %s sec" % (end - start))

    def testGetWelcome(self):
        """Test the 'getCouchWelcome' method"""
        resp = self.server.getCouchWelcome()
        self.assertEqual(resp["couchdb"], "Welcome")
        self.assertTrue("vendor" in resp)
        # For CouchDB 1.6.1, expected output is like:
        # {"couchdb":"Welcome","uuid":"2fe221d694085fa52409a9dc59d9f3f0","version":"1.6.1",
        # "vendor":{"version":"1.6.1","name":"The Apache Software Foundation"}}
        # Different output between CouchDB 1.6 vs CouchDB 3.1
        if resp["version"] != "1.6.1":
            self.assertTrue("features" in resp)

class CouchMonitorTest(unittest.TestCase):
    """
    Tests for the CouchMonitor class
    """

    def setUp(self):
        # Make an instance of the server
        self.couchUrl = os.getenv("COUCHURL", 'http://admin:password@localhost:5984')
        self.monitor = CouchMonitor(self.couchUrl)
        # Create a database, drop an existing one first
        self.dbNames = ['couchmonitor_unittest1', 'couchmonitor_unittest2']
        for dbName in self.dbNames:
            if dbName in self.monitor.couchServer.listDatabases():
                self.monitor.couchServer.deleteDatabase(dbName)

        for dbName in self.dbNames:
            self.monitor.couchServer.createDatabase(dbName)

        # define source/target endpoints
        self.sourceUrl = f"{self.couchUrl}/{self.dbNames[0]}"
        self.targetUrl = f"{self.couchUrl}/{self.dbNames[1]}"

    def tearDown(self):
        """Destroy the test databases before starting the next test"""
        for dbName in self.dbNames:
            self.monitor.couchServer.deleteDatabase(dbName)

    def testGetActiveTasks(self):
        """Test the 'getActiveTasks' method"""
        resp = self.monitor.getActiveTasks()
        # we haven't defined any task yet
        self.assertEqual(resp, [])

        # now define a replication task
        resp = self.monitor.couchServer.replicate(self.sourceUrl, self.targetUrl, sleepSecs=5)
        self.assertTrue(resp["ok"])
        time.sleep(1)
        # FIXME: this was supposed to be 1. To be investigated!!!
        self.assertEqual(len(self.monitor.getActiveTasks()), 0)

    def testGetSchedulerJobs(self):
        """
        Test the 'getSchedulerJobs' method, which is only meaningful
        for CouchDB version >= 3
        """
        resp = self.monitor.couchServer.getCouchWelcome()
        if resp["version"] == "1.6.1":
            return

        resp = self.monitor.getSchedulerJobs()
        self.assertEqual(resp["total_rows"], 0)
        self.assertEqual(resp["jobs"], [])

        # now define a replication task
        resp = self.monitor.couchServer.replicate(self.sourceUrl, self.targetUrl, sleepSecs=5)
        self.assertTrue(resp["ok"])
        # give it a couple of seconds for it to show up as active task
        time.sleep(1)
        resp = self.monitor.getSchedulerJobs()
        self.assertEqual(resp["total_rows"], 1)
        self.assertEqual(len(resp["jobs"]), 1)
        self.assertEqual(len(resp["jobs"]["database"]), "_replicator")

    def testGetSchedulerDocs(self):
        """
        Test the 'getSchedulerDocs' method, which is only meaningful
        for CouchDB version >= 3
        """
        resp = self.monitor.couchServer.getCouchWelcome()
        if resp["version"] == "1.6.1":
            return

        resp = self.monitor.getSchedulerDocs()
        self.assertEqual(resp["total_rows"], 0)
        self.assertEqual(resp["docs"], [])

        # now define a replication task
        resp = self.monitor.couchServer.replicate(self.sourceUrl, self.targetUrl, sleepSecs=5)
        self.assertTrue(resp["ok"])
        # give it a couple of seconds for it to show up as active task
        time.sleep(1)
        resp = self.monitor.getSchedulerDocs()
        self.assertEqual(resp["total_rows"], 1)
        self.assertEqual(len(resp["docs"]), 1)
        self.assertEqual(len(resp["docs"]["database"]), "_replicator")

    def testCheckCouchReplications(self):
        """Very basic tests for the 'checkCouchReplications' method"""
        resp = self.monitor.checkCouchReplications([])
        self.assertEqual(resp["status"], "ok")
        self.assertEqual(resp["error_message"], "")

    def testCheckReplicationState(self):
        """Very basic tests for the 'checkReplicationState' method"""
        resp = self.monitor.checkReplicationState()
        self.assertEqual(resp["status"], "ok")
        self.assertEqual(resp["error_message"], "")

    def testIsReplicationOK(self):
        """ Very basic test for the 'isReplicationOK' method """
        # define last update as of now - 15min
        myUpdateOn = int(time.time() - 1 * 60)
        checkPoint = 600000  # every 10min (in ms)
        testRepl = dict(updated_on=myUpdateOn, checkpoint_interval=checkPoint)
        resp = self.monitor.isReplicationOK(testRepl)
        self.assertTrue(resp)

        # now change last update to now - 2h to reproduce an error
        myUpdateOn = int(time.time() - 20 * 60)
        testRepl = dict(updated_on=myUpdateOn, checkpoint_interval=checkPoint)
        resp = self.monitor.isReplicationOK(testRepl)
        self.assertFalse(resp)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
Unittests for PSetTweaks module

"""
import unittest
import PSetTweaks.PSetTweak as PSetTweak

from WMQuality.TestInitCouchApp import TestInitCouchApp as TestInit

class PSetTweakTest(unittest.TestCase):
    """
    unittest for PSetTweak class

    """

    def setUp(self):
        """
        setUp some basic functions

        """
        self.testInit = TestInit(__file__)
        self.testDir  = self.testInit.generateWorkDir()
        return

    def tearDown(self):
        """
        tearDown the test framework

        """
        self.testInit.tearDownCouch()

    def testA(self):
        """init"""

        tweak = PSetTweak.PSetTweak()


    def testB(self):
        """adding attributes"""

        tweak = PSetTweak.PSetTweak()
        tweak.addParameter("process.source.logicalFileNames", ["/store/whatever"])
        tweak.addParameter("process.source.fileNames", ["file:/store/whatever"])


    def testC(self):
        """iterating"""
        tweak = PSetTweak.PSetTweak()
        tweak.addParameter("process.source.logicalFileNames", ["/store/whatever"])
        tweak.addParameter("process.source.fileNames", ["file:/store/whatever"])

        tweak.addParameter("process.module1.module2.module3.parameter1", ["file:/store/whatever", "file:/store/whatever2"])
        tweak.addParameter("process.module1.module2.module3.parameter2", ["file:/store/whatever"])
        tweak.addParameter("process.module1.module2.module3.parameter3", ["file:/store/whatever"])
        tweak.addParameter("process.module1.module2.module3.parameter4", 1234)


    def testD(self):
        """pythonise"""
        tweak = PSetTweak.PSetTweak()
        tweak.addParameter("process.source.logicalFileNames", ["/store/whatever"])
        tweak.addParameter("process.source.fileNames", ["file:/store/whatever"])
        tweak.addParameter("process.module1.module2.module3.parameter1", ["file:/store/whatever", "file:/store/whatever2"])
        tweak.addParameter("process.module1.module2.module3.parameter2", ["file:/store/whatever"])
        tweak.addParameter("process.module1.module2.module3.parameter3", ["file:/store/whatever"])
        tweak.addParameter("process.module1.module2.module3.parameter4", 1234)

        #print tweak.pythonise()

        persistFile = "%s/PSetTweak_persist.py" % self.testDir
        tweak.persist(persistFile, "python")

        tweak2 = PSetTweak.PSetTweak()
        tweak2.unpersist(persistFile)

        #print tweak2.pythonise()



    def testE(self):
        """jsonise"""
        tweak = PSetTweak.PSetTweak()
        tweak.addParameter("process.source.logicalFileNames", ["/store/whatever"])
        tweak.addParameter("process.source.fileNames", ["file:/store/whatever"])
        tweak.addParameter("process.module1.module2.module3.parameter1", ["file:/store/whatever", "file:/store/whatever2"])
        tweak.addParameter("process.module1.module2.module3.parameter2", ["file:/store/whatever"])
        tweak.addParameter("process.module1.module2.module3.parameter3", ["file:/store/whatever"])
        tweak.addParameter("process.module1.module2.module3.parameter4", 1234)

        #print tweak.jsonise()


        persistFile = "%s/PSetTweak_persist.json" % self.testDir
        tweak.persist(persistFile, "json")
        tweak2 = PSetTweak.PSetTweak()
        tweak2.unpersist(persistFile)

        #print tweak2.jsonise()


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
_GetFinishedSubscriptions_

Oracle implementation of Subscriptions.GetFinishedSubscriptions

Created on Aug 29, 2012

@author: dballest
"""

from WMCore.WMBS.MySQL.Subscriptions.GetFinishedSubscriptions import GetFinishedSubscriptions as MySQLFinishedSubscriptions

class GetFinishedSubscriptions(MySQLFinishedSubscriptions):
    pass



#!/usr/bin/env python
# encoding: utf-8
"""
CouchObject_t.py

Created by Dave Evans on 2010-07-29.
Copyright (c) 2010 Fermilab. All rights reserved.
"""

import unittest
import os

from WMCore.GroupUser.CouchObject import CouchObject

class CouchObject_t(unittest.TestCase):
    def setUp(self):
        self.database = "groupuser"
        self.url = os.getenv("COUCHURL", "http://127.0.0.1:5984")

    def testA(self):
        """test connection"""
        cObj = CouchObject()
        cObj.cdb_database = self.database
        cObj.cdb_url = self.url

        self.assertEqual(cObj.connected, False)

        cObj.connect()

        self.assertEqual(cObj.connected, True)


    def testB(self):
        """test create/drop of document"""

        class TestObject(CouchObject):
            def __init__(self):
                CouchObject.__init__(self)
                self.cdb_document_data = "testicle"
                self.setdefault("name", "noname")
                self.setdefault("Right", "Nut")
                self.setdefault("Left", "Nut")
            document_id = property(lambda x : "test-%s" % x['name'] )


        cObj = TestObject()
        cObj.setCouch(self.url, self.database)


        cObj['name'] = "CouchObjectUnitTest"
        cObj['data'] = {  "key1": "value1", "key2": "value2"}
        cObj['listOfStuff'] = [1,2,3,4,5,6,7]


        cObj.create()


        cObj.get()

        cObj2 = TestObject()
        cObj2.setCouch(self.url, self.database)
        cObj2['name'] = "CouchObjectUnitTest"
        cObj2.get()

        cObj.drop()



if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
_CountOpenBlocks_

Oracle implementation of DBSBuffer.CountOpenBlocks
"""

from __future__ import division
from WMComponent.DBS3Buffer.MySQL.CountOpenBlocks import CountOpenBlocks as MySQLCountOpenBlocks

class CountOpenBlocks(MySQLCountOpenBlocks):
    """
    _CountOpenBlocks_

    """
    pass



"""
Unit tests for MSPileup.py module

Author: Valentin Kuznetsov <vkuznet [AT] gmail [DOT] com>
"""

# system modules
import unittest

import cherrypy

# WMCore modules
from Utils.Timers import gmtimeSeconds
from WMQuality.Emulators.EmulatedUnitTestCase import EmulatedUnitTestCase
from WMCore.MicroService.Tools.Common import getMSLogger
from WMCore.MicroService.MSPileup.MSPileup import MSPileup
from WMCore.MicroService.MSPileup.MSPileupData import customDID
from WMCore.MicroService.MSPileup.DataStructs.MSPileupObj import MSPileupObj, schema


class MSPileupTest(EmulatedUnitTestCase):
    "Unit test for MSPileup module"
    def setUp(self):
        """
        set up unit test generic objects
        """
        self.logger = getMSLogger(False)
        super(MSPileupTest, self).setUp()
        cherrypy.request.user = "test"
        self.validRSEs = ['rse1', 'rse2']
        msConfig = {'reqmgr2Url': 'http://localhost',
                    'authz_key': '123',
                    'rucioAccount': 'wmcore_pileup',
                    'rucioUrl': 'http://cms-rucio-int.cern.ch',
                    'rucioAuthUrl': 'https://cms-rucio-auth-int.cern.ch',
                    'mongoDB': 'msPileupDB',
                    'mongoDBCollection': 'msPileupDBCollection',
                    'mongoDBServer': 'mongodb://localhost',
                    'mongoDBReplicaSet': '',
                    'mongoDBUser': None,
                    'mongoDBPassword': None,
                    'validRSEs': self.validRSEs,
                    'mockMongoDB': True}
        self.mgr = MSPileup(msConfig)

        pname = '/lksjdflksdjf/kljsdklfjsldfj/PREMIX'
        self.spec = {'pileupName': pname}

        expectedRSEs = self.validRSEs
        fullReplicas = 0
        campaigns = ['c1', 'c2']
        data = {
            'pileupName': pname,
            'pileupType': 'classic',
            'expectedRSEs': expectedRSEs,
            'currentRSEs': expectedRSEs,
            'fullReplicas': fullReplicas,
            'campaigns': campaigns,
            'containerFraction': 1.0,
            'replicationGrouping': "ALL",
            'active': True,
            'pileupSize': 0,
            'ruleIds': []}

        obj = MSPileupObj(data, validRSEs=self.validRSEs)
        for key in ['insertTime', 'lastUpdateTime', 'activatedOn', 'deactivatedOn']:
            self.assertNotEqual(obj.data[key], 0)
        self.assertEqual(obj.data['expectedRSEs'], expectedRSEs)
        self.assertEqual(obj.data['fullReplicas'], fullReplicas)
        self.assertEqual(obj.data['campaigns'], campaigns)
        self.doc = obj.getPileupData()

    def createDoc(self):
        """Creates a pileup object in the database"""
        return self.mgr.createPileup(self.doc)

    def testMSPileupStatus(self):
        """Test MSPileup status API"""
        res = self.mgr.status()
        self.assertEqual(res['error'], '')
        self.assertEqual(res['thread_id'], 'MainThread')

    def testMSPileupGet(self):
        """Test MSPileup createPileup and getPileup API"""
        self.assertEqual(len(self.createDoc()), 0)

        # get doc
        res = self.mgr.getPileup(**self.spec)
        self.assertEqual(len(res), 1)

        # get doc with one filter
        spec = self.spec.copy()
        spec['filters'] = ['currentRSEs']
        res = self.mgr.getPileup(**spec)
        self.assertEqual(len(res), 1)
        self.assertEqual(len(res[0].keys()), len(spec['filters']))

        for k in res[0]:
            self.assertIn(k, spec['filters'])

        # get doc with two filters
        spec = self.spec.copy()
        spec['filters'] = ['currentRSEs', 'containerFraction']
        res = self.mgr.getPileup(**spec)
        self.assertEqual(len(res), 1)
        self.assertEqual(len(res[0].keys()), len(spec['filters']))

        for k in res[0]:
            self.assertIn(k, spec['filters'])

        # get doc with empty string filter
        spec = self.spec.copy()
        spec['filters'] = ['']
        res = self.mgr.getPileup(**spec)
        self.assertEqual(len(res), 1)
        self.assertEqual(len(res[0].keys()), len(schema().keys()))

        # get doc with two filters with one empty string
        spec = self.spec.copy()
        spec['filters'] = ['currentRSEs', '']
        res = self.mgr.getPileup(**spec)
        self.assertEqual(len(res), 1)
        self.assertEqual(len(res[0].keys()), len(spec['filters']) - 1)

        for k in res[0]:
            self.assertIn(k, spec['filters'])

    def testMSPileupQuery(self):
        """Test MSPileup createPileup and queryDatabase API"""
        self.assertEqual(len(self.createDoc()), 0)

        # query doc
        projection = ["pileupType"]
        res = self.mgr.queryDatabase(self.spec, projection)
        self.assertEqual(len(res), 1)

    def testMSPileupUpdate(self):
        """Test MSPileup createPileup and updatePileup API"""
        self.assertEqual(len(self.createDoc()), 0)
        self.assertEqual(self.doc['pileupType'], "classic")

        # update doc
        doc = dict(self.doc)
        doc["pileupType"] = "premix"
        # add transition record as it now requires for update API
        customName = customDID(doc['pileupName'])
        trec = {'DN': 'localhost-test', 'containerFraction': 1, 'customDID': customName, 'updateTime': gmtimeSeconds()}
        doc.update({'transition': [trec]})

        res = self.mgr.updatePileup(doc)
        self.assertEqual(len(res), 0)

        # get doc
        res = self.mgr.getPileup(**self.spec)
        self.assertEqual(res[0]['pileupType'], "premix")

    def testMSPileupUpdateTransition(self):
        """Test MSPileup updatePileup API with transition logic"""
        self.assertEqual(len(self.createDoc()), 0)
        self.assertEqual(self.doc['pileupType'], "classic")

        # update doc
        doc = dict(self.doc)
        doc["pileupType"] = "premix"
        # add transition record as it now requires for update API
        # note: for first transition record we do not change custom name
        # see https://gist.github.com/amaltaro/b4f9bafc0b58c10092a0735c635538b5
        customName = doc['pileupName']
        trec = {'DN': 'localhost-test', 'containerFraction': 1, 'customDID': customName, 'updateTime': gmtimeSeconds()}
        doc.update({'transition': [trec]})

        res = self.mgr.updatePileup(doc)
        self.assertEqual(len(res), 0)
        self.logger.info("updatePileup %s", res)

        # get doc
        res = self.mgr.getPileup(**self.spec)
        self.assertEqual(res[0]['pileupType'], "premix")
        self.logger.info("getPileup %s", res)

        # now we'll perform check for transition records according to logic outlined in
        # https://gist.github.com/amaltaro/b4f9bafc0b58c10092a0735c635538b5
        # Please note: we can only change steps 1-5 since there is no MSPileupTasks is involved

        # add new spec with transition change
        spec = {'pileupName': doc['pileupName'], 'containerFraction': 0.5}
        res = self.mgr.updatePileup(spec)
        self.logger.info("updatePileup %s", res)
        self.assertEqual(len(res), 0)

        # get doc
        res = self.mgr.getPileup(**self.spec)
        self.assertEqual(res[0]['pileupType'], "premix")
        self.logger.info("getPileup %s", res)
        record = res[0]
        self.assertEqual(len(record['transition']), 2)
        self.assertEqual(record['customName'], '')

        # check transition records
        for idx, rec in enumerate(record['transition']):
            if idx == 0:
                self.assertEqual(rec['customDID'], doc['pileupName'])
                self.assertEqual(rec['containerFraction'], 1.0)
            elif idx == 1:
                self.assertEqual(rec['customDID'], doc['pileupName'] + '-V1')
                self.assertEqual(rec['containerFraction'], 1.0)

    def testMSPileupDelete(self):
        """Test MSPileup createPileup and deletePileup API"""
        self.assertEqual(len(self.createDoc()), 0)

        # delete doc
        res = self.mgr.deletePileup(self.spec)
        self.assertEqual(len(res), 0)

        # get doc - already deleted
        res = self.mgr.getPileup(**self.spec)
        self.assertEqual(res, [])


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
CouchFileset_t.py

Created by Dave Evans on 2010-10-05.
Copyright (c) 2010 Fermilab. All rights reserved.
"""

from builtins import range
import unittest
import random

from WMCore.ACDC.CouchCollection import CouchCollection
from WMCore.ACDC.CouchFileset import CouchFileset

from WMQuality.TestInitCouchApp import TestInitCouchApp
from WMCore.GroupUser.User import makeUser
from WMCore.Services.UUIDLib import makeUUID

from WMCore.DataStructs.File import File


class CouchFileset_t(unittest.TestCase):
    def setUp(self):
        """
        _setUp_

        Set up couch test environment.
        """
        self.testInit = TestInitCouchApp(__file__)
        self.testInit.setupCouch("wmcore-acdc-couchfileset", "GroupUser",
                                 "ACDC")
        self.owner = makeUser("DMWM", "evansde77", self.testInit.couchUrl,
                              self.testInit.couchDbName)
        self.owner.connect()
        self.owner.create()
        return

    def tearDown(self):
        """
        _tearDown_

        Clean up couch instance
        """
        self.testInit.tearDownCouch()

    def testDropCount(self):
        """
        _testDropCount_

        Verify that dropping a fileset and counting the files in a fileset works
        correctly.
        """
        testCollectionA = CouchCollection(database=self.testInit.couchDbName,
                                          url=self.testInit.couchUrl,
                                          name="Thunderstruck")
        testCollectionB = CouchCollection(database=self.testInit.couchDbName,
                                          url=self.testInit.couchUrl,
                                          name="StruckThunder")

        testFiles = []
        for i in range(5):
            testFile = File(lfn=makeUUID(), size=random.randint(1024, 4096),
                            events=random.randint(1024, 4096))
            testFiles.append(testFile)

        testFilesetA = CouchFileset(database=self.testInit.couchDbName,
                                    url=self.testInit.couchUrl,
                                    name="TestFilesetA")
        testFilesetB = CouchFileset(database=self.testInit.couchDbName,
                                    url=self.testInit.couchUrl,
                                    name="TestFilesetB")
        testFilesetC = CouchFileset(database=self.testInit.couchDbName,
                                    url=self.testInit.couchUrl,
                                    name="TestFilesetC")
        testCollectionA.addFileset(testFilesetA)
        testCollectionB.addFileset(testFilesetB)
        testCollectionB.addFileset(testFilesetC)
        testFilesetA.add(testFiles)
        testFilesetB.add(testFiles)
        testFilesetC.add(testFiles)

        testFilesetC.drop()

        testCollectionC = CouchCollection(database=self.testInit.couchDbName,
                                          url=self.testInit.couchUrl,
                                          name="StruckThunder")
        testCollectionC.populate()

        self.assertEqual(len(testCollectionC["filesets"]), 1,
                         "Error: There should be one fileset in this collection.")
        self.assertEqual(testCollectionC["filesets"][0].fileCount(), 5,
                         "Error: Wrong number of files in fileset.")

        testCollectionD = CouchCollection(database=self.testInit.couchDbName,
                                          url=self.testInit.couchUrl,
                                          name="Thunderstruck")
        testCollectionD.populate()

        self.assertEqual(len(testCollectionD["filesets"]), 1,
                         "Error: There should be one fileset in this collection.")
        self.assertEqual(testCollectionD["filesets"][0].fileCount(), 5,
                         "Error: Wrong number of files in fileset.")
        return

    def testListFiles(self):
        """
        _testListFiles_

        Verify that the files iterator works correctly.
        """
        testCollection = CouchCollection(database=self.testInit.couchDbName,
                                         url=self.testInit.couchUrl,
                                         name="Thunderstruck")
        testFileset = CouchFileset(database=self.testInit.couchDbName,
                                   url=self.testInit.couchUrl,
                                   name="TestFileset")
        testCollection.addFileset(testFileset)

        testFiles = {}
        for i in range(5):
            lfn = makeUUID()
            testFile = File(lfn=lfn, size=random.randint(1024, 4096),
                            events=random.randint(1024, 4096))
            testFiles[lfn] = testFile
            testFileset.add([testFile])

        for file in testFileset.listFiles():
            self.assertTrue(file["lfn"] in testFiles,
                            "Error: File missing.")
            self.assertEqual(file["events"], testFiles[file["lfn"]]["events"],
                             "Error: Wrong number of events.")
            self.assertEqual(file["size"], testFiles[file["lfn"]]["size"],
                             "Error: Wrong file size.")
        return

    def testFileset(self):
        """
        _testFileset_

        Verify that converting an ACDC fileset to a DataStructs fileset works
        correctly.
        """
        testCollection = CouchCollection(database=self.testInit.couchDbName,
                                         url=self.testInit.couchUrl,
                                         name="Thunderstruck")
        testFileset = CouchFileset(database=self.testInit.couchDbName,
                                   url=self.testInit.couchUrl,
                                   name="TestFileset")
        testCollection.addFileset(testFileset)

        testFiles = {}
        for i in range(5):
            lfn = makeUUID()
            testFile = File(lfn=lfn, size=random.randint(1024, 4096),
                            events=random.randint(1024, 4096))
            testFiles[lfn] = testFile
            testFileset.add([testFile])

        for file in testFileset.fileset().files:
            self.assertTrue(file["lfn"] in testFiles,
                            "Error: File missing.")
            self.assertEqual(file["events"], testFiles[file["lfn"]]["events"],
                             "Error: Wrong number of events.")
            self.assertEqual(file["size"], testFiles[file["lfn"]]["size"],
                             "Error: Wrong file size.")
        return


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# pylint: disable=E1101, C0103
# E1101: Use configuration files
# C0103: Different naming conventions apply for tests
"""
RetryManager test for module and the harness
"""
from __future__ import print_function

from builtins import range
import os
import os.path
import threading
import time
import unittest
from nose.plugins.attrib import attr

import WMCore.WMBase
from WMComponent.RetryManager.RetryManagerPoller import RetryManagerPoller
from WMCore.DAOFactory import DAOFactory
from WMCore.DataStructs.Run import Run
from WMCore.FwkJobReport.Report import Report
from WMCore.JobStateMachine.ChangeState import ChangeState
from WMCore.Services.UUIDLib import makeUUID
from WMCore.WMBS.File import File
from WMCore.WMBS.Fileset import Fileset
from WMCore.WMBS.Job import Job
from WMCore.WMBS.JobGroup import JobGroup
from WMCore.WMBS.Subscription import Subscription
from WMCore.WMBS.Workflow import Workflow
from WMQuality.Emulators import EmulatorSetup
from WMQuality.TestInitCouchApp import TestInitCouchApp as TestInit
from WMQuality.Emulators.EmulatedUnitTestCase import EmulatedUnitTestCase


class RetryManagerTest(EmulatedUnitTestCase):
    """
    TestCase for TestRetryManager module
    """

    def setUp(self):
        """
        setup for test.
        """
        super(RetryManagerTest, self).setUp()
        myThread = threading.currentThread()

        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection()
        self.testInit.setSchema(customModules=["WMCore.WMBS"],
                                useDefault=False)
        self.testInit.setupCouch("retry_manager_t/jobs", "JobDump")
        self.testInit.setupCouch("retry_manager_t/fwjrs", "FWJRDump")

        self.daofactory = DAOFactory(package="WMCore.WMBS",
                                     logger=myThread.logger,
                                     dbinterface=myThread.dbi)
        self.getJobs = self.daofactory(classname="Jobs.GetAllJobs")
        self.setJobTime = self.daofactory(classname="Jobs.SetStateTime")
        self.increaseRetry = self.daofactory(classname="Jobs.IncrementRetry")
        self.testDir = self.testInit.generateWorkDir()
        self.configFile = EmulatorSetup.setupWMAgentConfig()
        self.nJobs = 10
        return

    def tearDown(self):
        """
        Database deletion
        """
        self.testInit.clearDatabase()
        self.testInit.delWorkDir()
        self.testInit.tearDownCouch()
        EmulatorSetup.deleteConfig(self.configFile)
        return

    def getConfig(self):
        """
        _getConfig_

        """
        config = self.testInit.getConfiguration()
        self.testInit.generateWorkDir(config)

        # First the general stuff
        config.section_("General")
        config.General.workDir = os.getenv("TESTDIR", self.testDir)
        config.section_("CoreDatabase")
        config.CoreDatabase.connectUrl = os.getenv("DATABASE")
        config.CoreDatabase.socket = os.getenv("DBSOCK")

        config.component_("RetryManager")
        config.RetryManager.logLevel = 'DEBUG'
        config.RetryManager.namespace = 'WMComponent.RetryManager.RetryManager'
        config.RetryManager.pollInterval = 10
        # These are the cooloff times for the RetryManager, the times it waits
        # Before attempting resubmission
        config.RetryManager.section_("DefaultRetryAlgo")
        config.RetryManager.DefaultRetryAlgo.section_("default")
        config.RetryManager.DefaultRetryAlgo.default.coolOffTime = {'create': 120, 'submit': 120, 'job': 120}
        # Path to plugin directory
        config.RetryManager.pluginPath = 'WMComponent.RetryManager.PlugIns'
        config.RetryManager.WMCoreBase = WMCore.WMBase.getWMBASE()
        config.RetryManager.componentDir = os.path.join(os.getcwd(), 'Components')

        # ErrorHandler
        # Not essential, but useful for ProcessingAlgo
        config.component_("ErrorHandler")
        config.ErrorHandler.maxRetries = 5

        # JobStateMachine
        config.component_('JobStateMachine')
        config.JobStateMachine.couchurl = os.getenv('COUCHURL', None)
        config.JobStateMachine.couchDBName = "retry_manager_t"

        return config

    def createTestJobGroup(self, nJobs, subType="Processing", retryOnce=False):
        """
        _createTestJobGroup_

        Creates a group of several jobs
        """
        testWorkflow = Workflow(spec="spec.xml", owner="Simon",
                                name=makeUUID(), task="Test")
        testWorkflow.create()

        testWMBSFileset = Fileset(name="TestFileset")
        testWMBSFileset.create()
        testSubscription = Subscription(fileset=testWMBSFileset,
                                        workflow=testWorkflow,
                                        type=subType)
        testSubscription.create()

        testJobGroup = JobGroup(subscription=testSubscription)
        testJobGroup.create()

        testFileA = File(lfn="/this/is/a/lfnA", size=1024, events=10)
        testFileA.addRun(Run(10, *[12312]))
        testFileA.setLocation('malpaquet')

        testFileB = File(lfn="/this/is/a/lfnB", size=1024, events=10)
        testFileB.addRun(Run(10, *[12312]))
        testFileA.setLocation('malpaquet')
        testFileA.create()
        testFileB.create()

        for _ in range(0, nJobs):
            testJob = Job(name=makeUUID())
            testJob.addFile(testFileA)
            testJob.addFile(testFileB)
            testJob['cache_dir'] = os.path.join(self.testDir, testJob['name'])
            os.mkdir(testJob['cache_dir'])
            testJobGroup.add(testJob)

        testJobGroup.commit()
        if retryOnce:
            self.increaseRetry.execute(testJobGroup.jobs)

        return testJobGroup

    def testA_Create(self):
        """
        WMComponent_t.RetryManager_t.RetryManager_t:testCreate()

        Mimics creation of component and test jobs failed in create stage.
        """
        testJobGroup = self.createTestJobGroup(nJobs=self.nJobs)

        config = self.getConfig()
        changer = ChangeState(config)
        changer.propagate(testJobGroup.jobs, 'createfailed', 'new')
        changer.propagate(testJobGroup.jobs, 'createcooloff', 'createfailed')

        idList = self.getJobs.execute(state='CreateCooloff')
        self.assertEqual(len(idList), self.nJobs)

        testRetryManager = RetryManagerPoller(config)
        testRetryManager.setup(None)

        for job in testJobGroup.jobs:
            self.setJobTime.execute(jobID=job["id"],
                                    stateTime=int(time.time()) - 50)

        testRetryManager.algorithm(None)
        idList = self.getJobs.execute(state='CreateCooloff')
        self.assertEqual(len(idList), self.nJobs)

        for job in testJobGroup.jobs:
            self.setJobTime.execute(jobID=job["id"],
                                    stateTime=int(time.time()) - 150)

        testRetryManager.algorithm(None)

        idList = self.getJobs.execute(state='CreateCooloff')
        self.assertEqual(len(idList), 0)

        idList = self.getJobs.execute(state='Created')
        self.assertEqual(len(idList), self.nJobs)
        return

    def testB_Submit(self):
        """
        WMComponent_t.RetryManager_t.RetryManager_t:testSubmit()

        Mimics creation of component and test jobs failed in create stage.
        """
        testJobGroup = self.createTestJobGroup(nJobs=self.nJobs)

        config = self.getConfig()
        changer = ChangeState(config)
        changer.propagate(testJobGroup.jobs, 'created', 'new')
        changer.propagate(testJobGroup.jobs, 'submitfailed', 'created')
        changer.propagate(testJobGroup.jobs, 'submitcooloff', 'submitfailed')

        idList = self.getJobs.execute(state='SubmitCooloff')
        self.assertEqual(len(idList), self.nJobs)

        testRetryManager = RetryManagerPoller(config)
        testRetryManager.setup(None)

        for job in testJobGroup.jobs:
            self.setJobTime.execute(jobID=job["id"],
                                    stateTime=int(time.time()) - 50)

        testRetryManager.algorithm(None)

        idList = self.getJobs.execute(state='SubmitCooloff')
        self.assertEqual(len(idList), self.nJobs)

        for job in testJobGroup.jobs:
            self.setJobTime.execute(jobID=job["id"],
                                    stateTime=int(time.time()) - 150)

        testRetryManager.algorithm(None)

        idList = self.getJobs.execute(state='SubmitCooloff')
        self.assertEqual(len(idList), 0)

        idList = self.getJobs.execute(state='Created')
        self.assertEqual(len(idList), self.nJobs)
        return

    def testC_Job(self):
        """
        WMComponent_t.RetryManager_t.RetryManager_t:testJob()

        Mimics creation of component and test jobs failed in create stage.
        """
        testJobGroup = self.createTestJobGroup(nJobs=self.nJobs)

        config = self.getConfig()
        changer = ChangeState(config)
        changer.propagate(testJobGroup.jobs, 'created', 'new')
        changer.propagate(testJobGroup.jobs, 'executing', 'created')
        changer.propagate(testJobGroup.jobs, 'complete', 'executing')
        changer.propagate(testJobGroup.jobs, 'jobfailed', 'complete')
        changer.propagate(testJobGroup.jobs, 'jobcooloff', 'jobfailed')

        idList = self.getJobs.execute(state='JobCooloff')
        self.assertEqual(len(idList), self.nJobs)

        testRetryManager = RetryManagerPoller(config)
        testRetryManager.setup(None)

        for job in testJobGroup.jobs:
            self.setJobTime.execute(jobID=job["id"],
                                    stateTime=int(time.time()) - 50)

        testRetryManager.algorithm(None)

        idList = self.getJobs.execute(state='JobCooloff')
        self.assertEqual(len(idList), self.nJobs)

        for job in testJobGroup.jobs:
            self.setJobTime.execute(jobID=job["id"],
                                    stateTime=int(time.time()) - 150)

        testRetryManager.algorithm(None)

        idList = self.getJobs.execute(state='JobCooloff')
        self.assertEqual(len(idList), 0)

        idList = self.getJobs.execute(state='Created')
        self.assertEqual(len(idList), self.nJobs)
        return

    def testD_SquaredAlgo(self):
        """
        _testSquaredAlgo_

        Test the squared algorithm to make sure it loads and works
        """

        testJobGroup = self.createTestJobGroup(nJobs=self.nJobs)

        config = self.getConfig()
        config.RetryManager.plugins = {'Processing': 'SquaredAlgo'}
        config.RetryManager.section_("SquaredAlgo")
        config.RetryManager.SquaredAlgo.section_("Processing")
        config.RetryManager.SquaredAlgo.Processing.coolOffTime = {'create': 10, 'submit': 10, 'job': 10}
        changer = ChangeState(config)
        changer.propagate(testJobGroup.jobs, 'created', 'new')
        changer.propagate(testJobGroup.jobs, 'submitfailed', 'created')
        changer.propagate(testJobGroup.jobs, 'submitcooloff', 'submitfailed')
        changer.propagate(testJobGroup.jobs, 'created', 'submitcooloff')
        changer.propagate(testJobGroup.jobs, 'submitfailed', 'created')
        changer.propagate(testJobGroup.jobs, 'submitcooloff', 'submitfailed')

        idList = self.getJobs.execute(state='SubmitCooloff')
        self.assertEqual(len(idList), self.nJobs)

        testRetryManager = RetryManagerPoller(config)
        testRetryManager.setup(None)

        for job in testJobGroup.jobs:
            self.setJobTime.execute(jobID=job["id"],
                                    stateTime=int(time.time()) - 5)

        testRetryManager.algorithm(None)
        idList = self.getJobs.execute(state='SubmitCooloff')
        self.assertEqual(len(idList), self.nJobs)

        for job in testJobGroup.jobs:
            self.setJobTime.execute(jobID=job["id"],
                                    stateTime=int(time.time()) - 12)

        testRetryManager.algorithm(None)

        idList = self.getJobs.execute(state='SubmitCooloff')
        self.assertEqual(len(idList), 0)

        idList = self.getJobs.execute(state='Created')
        self.assertEqual(len(idList), self.nJobs)

    def testE_ExponentialAlgo(self):
        """
        _testExponentialAlgo_

        Test the exponential algorithm to make sure it loads and works
        """

        testJobGroup = self.createTestJobGroup(nJobs=self.nJobs)

        config = self.getConfig()
        config.RetryManager.plugins = {'Processing': 'ExponentialAlgo'}
        config.RetryManager.section_("ExponentialAlgo")
        config.RetryManager.ExponentialAlgo.section_("Processing")
        config.RetryManager.ExponentialAlgo.Processing.coolOffTime = {'create': 10, 'submit': 10, 'job': 10}
        changer = ChangeState(config)
        changer.propagate(testJobGroup.jobs, 'created', 'new')
        changer.propagate(testJobGroup.jobs, 'submitfailed', 'created')
        changer.propagate(testJobGroup.jobs, 'submitcooloff', 'submitfailed')
        changer.propagate(testJobGroup.jobs, 'created', 'submitcooloff')
        changer.propagate(testJobGroup.jobs, 'submitfailed', 'created')
        changer.propagate(testJobGroup.jobs, 'submitcooloff', 'submitfailed')

        idList = self.getJobs.execute(state='SubmitCooloff')
        self.assertEqual(len(idList), self.nJobs)

        testRetryManager = RetryManagerPoller(config)
        testRetryManager.setup(None)

        for job in testJobGroup.jobs:
            self.setJobTime.execute(jobID=job["id"],
                                    stateTime=int(time.time()) - 5)

        testRetryManager.algorithm(None)
        idList = self.getJobs.execute(state='SubmitCooloff')
        self.assertEqual(len(idList), self.nJobs)

        for job in testJobGroup.jobs:
            self.setJobTime.execute(jobID=job["id"],
                                    stateTime=int(time.time()) - 12)

        testRetryManager.algorithm(None)

        idList = self.getJobs.execute(state='SubmitCooloff')
        self.assertEqual(len(idList), 0)

        idList = self.getJobs.execute(state='Created')
        self.assertEqual(len(idList), self.nJobs)

    def testF_LinearAlgo(self):
        """
        _testLinearAlgo_

        Test the linear algorithm to make sure it loads and works
        """

        testJobGroup = self.createTestJobGroup(nJobs=self.nJobs)

        config = self.getConfig()
        config.RetryManager.plugins = {'Processing': 'LinearAlgo'}
        config.RetryManager.section_("LinearAlgo")
        config.RetryManager.LinearAlgo.section_("Processing")
        config.RetryManager.LinearAlgo.Processing.coolOffTime = {'create': 10, 'submit': 10, 'job': 10}
        changer = ChangeState(config)
        changer.propagate(testJobGroup.jobs, 'created', 'new')
        changer.propagate(testJobGroup.jobs, 'submitfailed', 'created')
        changer.propagate(testJobGroup.jobs, 'submitcooloff', 'submitfailed')
        changer.propagate(testJobGroup.jobs, 'created', 'submitcooloff')
        changer.propagate(testJobGroup.jobs, 'submitfailed', 'created')
        changer.propagate(testJobGroup.jobs, 'submitcooloff', 'submitfailed')

        idList = self.getJobs.execute(state='SubmitCooloff')
        self.assertEqual(len(idList), self.nJobs)

        testRetryManager = RetryManagerPoller(config)
        testRetryManager.setup(None)

        for job in testJobGroup.jobs:
            self.setJobTime.execute(jobID=job["id"],
                                    stateTime=int(time.time()) - 5)

        testRetryManager.algorithm(None)
        idList = self.getJobs.execute(state='SubmitCooloff')
        self.assertEqual(len(idList), self.nJobs)

        for job in testJobGroup.jobs:
            self.setJobTime.execute(jobID=job["id"],
                                    stateTime=int(time.time()) - 12)

        testRetryManager.algorithm(None)

        idList = self.getJobs.execute(state='SubmitCooloff')
        self.assertEqual(len(idList), 0)

        idList = self.getJobs.execute(state='Created')
        self.assertEqual(len(idList), self.nJobs)

        return

    def testG_ProcessingAlgo(self):
        """
        _ProcessingAlgo_

        Test for the ProcessingAlgo Prototype
        """

        testJobGroup = self.createTestJobGroup(nJobs=self.nJobs)

        config = self.getConfig()
        config.RetryManager.plugins = {'Processing': 'ProcessingAlgo'}
        config.RetryManager.section_("ProcessingAlgo")
        config.RetryManager.ProcessingAlgo.section_("default")
        config.RetryManager.ProcessingAlgo.default.coolOffTime = {'create': 10, 'submit': 10, 'job': 10}
        changer = ChangeState(config)
        fwjrPath = os.path.join(WMCore.WMBase.getTestBase(),
                                "WMComponent_t/JobAccountant_t",
                                "fwjrs/badBackfillJobReport.pkl")
        report = Report()
        report.load(fwjrPath)
        for job in testJobGroup.jobs:
            job['fwjr'] = report
            job['retry_count'] = 0
            report.save(os.path.join(job['cache_dir'], "Report.%i.pkl" % job['retry_count']))
        changer.propagate(testJobGroup.jobs, 'created', 'new')
        changer.propagate(testJobGroup.jobs, 'executing', 'created')
        changer.propagate(testJobGroup.jobs, 'complete', 'executing')
        changer.propagate(testJobGroup.jobs, 'jobfailed', 'complete')
        changer.propagate(testJobGroup.jobs, 'jobcooloff', 'jobfailed')

        testRetryManager = RetryManagerPoller(config)
        testRetryManager.algorithm()

        idList = self.getJobs.execute(state='Created')
        self.assertEqual(len(idList), self.nJobs)

        changer.propagate(testJobGroup.jobs, 'executing', 'created')
        changer.propagate(testJobGroup.jobs, 'complete', 'executing')
        changer.propagate(testJobGroup.jobs, 'jobfailed', 'complete')
        changer.propagate(testJobGroup.jobs, 'jobcooloff', 'jobfailed')

        for job in testJobGroup.jobs:
            j = Job(id=job['id'])
            j.load()
            self.assertEqual(j['retry_count'], 1)
            report.save(os.path.join(j['cache_dir'], "Report.%i.pkl" % j['retry_count']))

        config.RetryManager.ProcessingAlgo.default.OneMoreErrorCodes = [8020]
        testRetryManager2 = RetryManagerPoller(config)
        testRetryManager2.algorithm()

        idList = self.getJobs.execute(state='Created')
        self.assertEqual(len(idList), self.nJobs)

        for job in testJobGroup.jobs:
            j = Job(id=job['id'])
            j.load()
            self.assertEqual(j['retry_count'], 5)

        # Now test timeout
        testJobGroup2 = self.createTestJobGroup(nJobs=self.nJobs)

        # Cycle jobs
        for job in testJobGroup2.jobs:
            job['fwjr'] = report
            job['retry_count'] = 0
            report.save(os.path.join(job['cache_dir'], "Report.%i.pkl" % job['retry_count']))
        changer.propagate(testJobGroup2.jobs, 'created', 'new')
        changer.propagate(testJobGroup2.jobs, 'executing', 'created')
        changer.propagate(testJobGroup2.jobs, 'complete', 'executing')
        changer.propagate(testJobGroup2.jobs, 'jobfailed', 'complete')
        changer.propagate(testJobGroup2.jobs, 'jobcooloff', 'jobfailed')

        for job in testJobGroup2.jobs:
            j = Job(id=job['id'])
            j.load()
            self.assertEqual(j['retry_count'], 0)

        config.RetryManager.ProcessingAlgo.default.OneMoreErrorCodes = []
        config.RetryManager.ProcessingAlgo.default.MaxRunTime = 1
        testRetryManager3 = RetryManagerPoller(config)
        testRetryManager3.algorithm()

        idList = self.getJobs.execute(state='Created')
        self.assertEqual(len(idList), self.nJobs * 2)

        for job in testJobGroup2.jobs:
            j = Job(id=job['id'])
            j.load()
            self.assertEqual(j['retry_count'], 5)

        return

    def testH_PauseAlgo(self):
        """
        _testH_PauseAlgo_

        Test the pause algorithm, note that given pauseCount = n, the
        job will run first n + 1 times before being paused.
        After that it will be paused each n times
        """

        testJobGroup = self.createTestJobGroup(nJobs=self.nJobs)

        # adding a 2nd job group
        testJobGroup2 = self.createTestJobGroup(nJobs=self.nJobs)

        config = self.getConfig()
        config.RetryManager.plugins = {'Processing': 'PauseAlgo'}
        config.RetryManager.section_("PauseAlgo")
        config.RetryManager.PauseAlgo.section_("Processing")
        config.RetryManager.PauseAlgo.Processing.coolOffTime = {'create': 20, 'submit': 20, 'job': 20}
        config.RetryManager.PauseAlgo.Processing.pauseCount = 2
        changer = ChangeState(config)
        changer.propagate(testJobGroup.jobs, 'created', 'new')
        changer.propagate(testJobGroup.jobs, 'executing', 'created')
        changer.propagate(testJobGroup.jobs, 'jobfailed', 'executing')
        changer.propagate(testJobGroup.jobs, 'jobcooloff', 'jobfailed')
        changer.propagate(testJobGroup.jobs, 'created', 'jobcooloff')
        changer.propagate(testJobGroup.jobs, 'executing', 'created')
        changer.propagate(testJobGroup.jobs, 'jobfailed', 'executing')
        changer.propagate(testJobGroup.jobs, 'jobcooloff', 'jobfailed')

        testRetryManager = RetryManagerPoller(config)
        testRetryManager.setup(None)

        report = Report()

        # Making sure that jobs are not created ahead of time
        for job in testJobGroup.jobs:
            self.setJobTime.execute(jobID=job["id"],
                                    stateTime=int(time.time()) - 15)
        testRetryManager.algorithm(None)
        idList = self.getJobs.execute(state='JobCoolOff')
        self.assertEqual(len(idList), self.nJobs)

        # Giving time so they can be retried
        for job in testJobGroup.jobs:
            self.setJobTime.execute(jobID=job["id"],
                                    stateTime=int(time.time()) - 25)

        # Make sure that the plugin allowed them to go back to created state
        testRetryManager.algorithm(None)
        idList = self.getJobs.execute(state='created')
        self.assertEqual(len(idList), self.nJobs)

        # Fail them out again
        changer.propagate(testJobGroup.jobs, 'executing', 'created')
        changer.propagate(testJobGroup.jobs, 'jobfailed', 'executing')
        changer.propagate(testJobGroup.jobs, 'jobcooloff', 'jobfailed')

        # Make sure that no change happens before timeout
        for job in testJobGroup.jobs:
            self.setJobTime.execute(jobID=job["id"],
                                    stateTime=int(time.time()) - 75)
        testRetryManager.algorithm(None)
        idList = self.getJobs.execute(state='JobCoolOff')
        self.assertEqual(len(idList), self.nJobs)

        # Giving time so they can be paused
        for job in testJobGroup.jobs:
            self.setJobTime.execute(jobID=job["id"],
                                    stateTime=int(time.time()) - 85)

        # Make sure that the plugin pauses them
        testRetryManager.algorithm(None)
        idList = self.getJobs.execute(state='jobpaused')
        self.assertEqual(len(idList), self.nJobs)

        # Emulating ops retrying the job
        changer.propagate(testJobGroup.jobs, 'created', 'jobpaused')

        # Making sure it did the right thing
        idList = self.getJobs.execute(state='created')
        self.assertEqual(len(idList), self.nJobs)

        # Fail them out again
        changer.propagate(testJobGroup.jobs, 'executing', 'created')
        changer.propagate(testJobGroup.jobs, 'jobfailed', 'executing')
        changer.propagate(testJobGroup.jobs, 'jobfailed', 'executing')
        changer.propagate(testJobGroup.jobs, 'jobcooloff', 'jobfailed')

        for job in testJobGroup.jobs:
            self.setJobTime.execute(jobID=job["id"],
                                    stateTime=int(time.time()) - 175)
        testRetryManager.algorithm(None)
        idList = self.getJobs.execute(state='JobCoolOff')
        self.assertEqual(len(idList), self.nJobs)

        # Giving time so they can be retried
        for job in testJobGroup.jobs:
            self.setJobTime.execute(jobID=job["id"],
                                    stateTime=int(time.time()) - 185)

        # Make sure that the plugin allowed them to go back to created state
        testRetryManager.algorithm(None)
        idList = self.getJobs.execute(state='created')
        self.assertEqual(len(idList), self.nJobs)

        # Fail them out again
        changer.propagate(testJobGroup.jobs, 'executing', 'created')
        changer.propagate(testJobGroup.jobs, 'jobfailed', 'executing')
        changer.propagate(testJobGroup.jobs, 'jobfailed', 'executing')
        changer.propagate(testJobGroup.jobs, 'jobcooloff', 'jobfailed')

        for job in testJobGroup.jobs:
            self.setJobTime.execute(jobID=job["id"],
                                    stateTime=int(time.time()) - 315)
        testRetryManager.algorithm(None)
        idList = self.getJobs.execute(state='jobcooloff')
        self.assertEqual(len(idList), self.nJobs)

        for job in testJobGroup.jobs:
            self.setJobTime.execute(jobID=job["id"],
                                    stateTime=int(time.time()) - 325)

        # Make sure that the plugin allowed them to go back to created state
        testRetryManager.algorithm(None)
        idList = self.getJobs.execute(state='jobpaused')
        self.assertEqual(len(idList), self.nJobs)

        # a configurable retry count per job type {jobExitCodeA: pauseCountB}
        config.RetryManager.PauseAlgo.Processing.retryErrorCodes = {8020: 1, 12345: 1, 5555: 2}

        testRetryManager2 = RetryManagerPoller(config)
        testRetryManager2.algorithm()

        fwjrPath = os.path.join(WMCore.WMBase.getTestBase(),
                                "WMComponent_t/JobAccountant_t",
                                "fwjrs/badBackfillJobReport.pkl")

        report.load(fwjrPath)
        for job in testJobGroup2.jobs:
            job['fwjr'] = report
            job['retry_count'] = 0
            report.save(os.path.join(job['cache_dir'], "Report.%i.pkl" % job['retry_count']))

        # fail the jobs
        changer.propagate(testJobGroup2.jobs, 'created', 'new')
        changer.propagate(testJobGroup2.jobs, 'executing', 'created')
        changer.propagate(testJobGroup2.jobs, 'complete', 'executing')
        changer.propagate(testJobGroup2.jobs, 'jobfailed', 'complete')
        changer.propagate(testJobGroup2.jobs, 'jobcooloff', 'jobfailed')

        # Giving time so they can be paused
        for job in testJobGroup2.jobs:
            self.setJobTime.execute(jobID=job["id"],
                                    stateTime=int(time.time()) - 85)

        # Make sure that the plugin sent those jobs to the next state:
        testRetryManager2.algorithm()
        # job exit code is 8020, so it is supposed to be retried one time.
        # Meaning, that here we should have 10 jobs (from the first part of the test) in jobpaused
        # and 10 jobs in created state

        idList = self.getJobs.execute(state='created')
        self.assertEqual(len(idList), self.nJobs)

        idList2 = self.getJobs.execute(state='jobpaused')
        self.assertEqual(len(idList2), self.nJobs)

        # save a second job report - with a retry count = 1
        for job in testJobGroup2.jobs:
            j = Job(id=job['id'])
            j.load()
            j['retry_count'] = 1
            self.assertEqual(j['retry_count'], 1)
            report.save(os.path.join(j['cache_dir'], "Report.%i.pkl" % j['retry_count']))

        # Fail them out again
        changer.propagate(testJobGroup2.jobs, 'executing', 'created')
        changer.propagate(testJobGroup2.jobs, 'complete', 'executing')
        changer.propagate(testJobGroup2.jobs, 'jobfailed', 'complete')
        changer.propagate(testJobGroup2.jobs, 'jobcooloff', 'jobfailed')

        for job in testJobGroup2.jobs:
            self.setJobTime.execute(jobID=job["id"],
                                    stateTime=int(time.time()) - 175)

        # not sure if this check is needed:
        idList = self.getJobs.execute(state='jobcooloff')
        self.assertEqual(len(idList), self.nJobs)

        # Giving time so they can be paused
        for job in testJobGroup2.jobs:
            self.setJobTime.execute(jobID=job["id"],
                                    stateTime=int(time.time()) - 85)

        # Make sure that the plugin sent those jobs to paused state:
        testRetryManager2.algorithm(None)
        idList = self.getJobs.execute(state='jobpaused')
        # And again, in total, there should be 10+10=20 jobs in jobpaused
        self.assertEqual(len(idList), self.nJobs * 2)

        return

    def testI_MultipleJobTypes(self):
        """
        _testI_MultipleJobTypes_

        Check that we can configure different retry algorithms for different
        job types, including a default for nonspecified types.
        Also check that two job types can share the same retry algorithm
        but with different parameters
        """

        # Let's create 4 job groups
        processingJobGroup = self.createTestJobGroup(nJobs=10,
                                                     retryOnce=True)
        productionJobGroup = self.createTestJobGroup(nJobs=15,
                                                     subType="Production",
                                                     retryOnce=True)
        mergeJobGroup = self.createTestJobGroup(nJobs=20,
                                                subType="Merge",
                                                retryOnce=True)
        skimJobGroup = self.createTestJobGroup(nJobs=5,
                                               subType="Skim",
                                               retryOnce=True)

        # Set an adequate config
        # Processing jobs get the PauseAlgo with pauseCount 4
        # Production jobs get the ExponentialAlgo
        # Merge jobs get the PauseAlgo but with pauseCount 2 which is the default
        # Skim jobs are not configured, so they get the default SquaredAlgo
        config = self.getConfig()
        config.RetryManager.plugins = {'Processing': 'PauseAlgo',
                                       'Production': 'ExponentialAlgo',
                                       'Merge': 'PauseAlgo',
                                       'default': 'SquaredAlgo'}
        config.RetryManager.section_("PauseAlgo")
        config.RetryManager.PauseAlgo.section_("Processing")
        config.RetryManager.PauseAlgo.Processing.coolOffTime = {'create': 30, 'submit': 30, 'job': 30}
        config.RetryManager.PauseAlgo.Processing.pauseCount = 4
        config.RetryManager.PauseAlgo.section_("default")
        config.RetryManager.PauseAlgo.default.coolOffTime = {'create': 60, 'submit': 60, 'job': 60}
        config.RetryManager.PauseAlgo.default.pauseCount = 2
        config.RetryManager.section_("ExponentialAlgo")
        config.RetryManager.ExponentialAlgo.section_("Production")
        config.RetryManager.ExponentialAlgo.Production.coolOffTime = {'create': 30, 'submit': 30, 'job': 30}
        config.RetryManager.ExponentialAlgo.section_("default")
        config.RetryManager.ExponentialAlgo.default.coolOffTime = {'create': 60, 'submit': 60, 'job': 60}
        config.RetryManager.section_("SquaredAlgo")
        config.RetryManager.SquaredAlgo.section_("Skim")
        config.RetryManager.SquaredAlgo.Skim.coolOffTime = {'create': 30, 'submit': 30, 'job': 30}
        config.RetryManager.SquaredAlgo.section_("default")
        config.RetryManager.SquaredAlgo.default.coolOffTime = {'create': 60, 'submit': 60, 'job': 60}

        # Start the state changer and RetryManager
        changer = ChangeState(config)
        testRetryManager = RetryManagerPoller(config)
        testRetryManager.setup(None)

        # Create the jobs for the first time
        changer.propagate(processingJobGroup.jobs, 'created', 'new')

        # Let's start with the processing jobs and the pauseAlgo
        for count in range(1, 5):
            # Fail the jobs
            changer.propagate(processingJobGroup.jobs, 'executing', 'created')
            changer.propagate(processingJobGroup.jobs, 'jobfailed', 'executing')
            changer.propagate(processingJobGroup.jobs, 'jobcooloff', 'jobfailed')

            # Check  that the cooloff time is strictly enforced
            # First a job time just below the cooloff time
            for job in processingJobGroup.jobs:
                self.setJobTime.execute(jobID=job["id"],
                                        stateTime=int(time.time()) - 30 * pow(count, 2) + 5)
            testRetryManager.algorithm(None)
            idList = self.getJobs.execute(state='JobCoolOff')
            self.assertEqual(len(idList), len(processingJobGroup.jobs),
                             "Jobs went into cooloff without the proper timing")

            # Now above the cooloff time
            for job in processingJobGroup.jobs:
                self.setJobTime.execute(jobID=job["id"],
                                        stateTime=int(time.time()) - 30 * pow(count, 2) - 5)
            testRetryManager.algorithm(None)

            # Make sure the jobs get created again or go to paused
            if count < 4:
                idList = self.getJobs.execute(state='created')
            else:
                idList = self.getJobs.execute(state='jobpaused')
            self.assertEqual(len(idList), len(processingJobGroup.jobs),
                             "Jobs didn't change state correctly")

        # Unpause them so they don't interfere with subsequent tests
        changer.propagate(processingJobGroup.jobs, 'created', 'jobpaused')
        changer.propagate(processingJobGroup.jobs, 'executing', 'created')

        # Now the production jobs and the exponential algo
        changer.propagate(productionJobGroup.jobs, 'created', 'new')

        for count in range(1, 3):
            changer.propagate(productionJobGroup.jobs, 'executing', 'created')
            changer.propagate(productionJobGroup.jobs, 'jobfailed', 'executing')
            changer.propagate(productionJobGroup.jobs, 'jobcooloff', 'jobfailed')

            for job in productionJobGroup.jobs:
                self.setJobTime.execute(jobID=job["id"],
                                        stateTime=int(time.time()) - pow(30, count) + 5)
            testRetryManager.algorithm(None)
            idList = self.getJobs.execute(state='JobCoolOff')
            self.assertEqual(len(idList), len(productionJobGroup.jobs),
                             "Jobs went into cooloff without the proper timing")
            for job in productionJobGroup.jobs:
                self.setJobTime.execute(jobID=job["id"],
                                        stateTime=int(time.time()) - pow(30, count) - 5)
            testRetryManager.algorithm(None)

            idList = self.getJobs.execute(state='created')
            self.assertEqual(len(idList), len(productionJobGroup.jobs),
                             "Jobs didn't change state correctly")

        # Send them to executing
        changer.propagate(productionJobGroup.jobs, 'executing', 'created')

        # Now the merge jobs and the paused algo with different parameters
        changer.propagate(mergeJobGroup.jobs, 'created', 'new')

        for count in range(1, 3):
            changer.propagate(mergeJobGroup.jobs, 'executing', 'created')
            changer.propagate(mergeJobGroup.jobs, 'jobfailed', 'executing')
            changer.propagate(mergeJobGroup.jobs, 'jobcooloff', 'jobfailed')

            for job in mergeJobGroup.jobs:
                self.setJobTime.execute(jobID=job["id"],
                                        stateTime=int(time.time()) - 30 * pow(count, 2) - 5)
            testRetryManager.algorithm(None)
            idList = self.getJobs.execute(state='JobCoolOff')
            self.assertEqual(len(idList), len(mergeJobGroup.jobs),
                             "Jobs went into cooloff without the proper timing")

            for job in mergeJobGroup.jobs:
                self.setJobTime.execute(jobID=job["id"],
                                        stateTime=int(time.time()) - 60 * pow(count, 2) - 5)
            testRetryManager.algorithm(None)

            if count < 2:
                idList = self.getJobs.execute(state='created')
            else:
                idList = self.getJobs.execute(state='jobpaused')
            self.assertEqual(len(idList), len(mergeJobGroup.jobs),
                             "Jobs didn't change state correctly")

        # Send them to executing
        changer.propagate(mergeJobGroup.jobs, 'created', 'jobpaused')
        changer.propagate(mergeJobGroup.jobs, 'executing', 'created')

        # Now the skim jobs and the squared algo
        changer.propagate(skimJobGroup.jobs, 'created', 'new')

        for count in range(1, 3):
            changer.propagate(skimJobGroup.jobs, 'executing', 'created')
            changer.propagate(skimJobGroup.jobs, 'jobfailed', 'executing')
            changer.propagate(skimJobGroup.jobs, 'jobcooloff', 'jobfailed')

            for job in skimJobGroup.jobs:
                self.setJobTime.execute(jobID=job["id"],
                                        stateTime=int(time.time()) - 30 * pow(count, 2) + 5)
            testRetryManager.algorithm(None)
            idList = self.getJobs.execute(state='JobCoolOff')
            self.assertEqual(len(idList), len(skimJobGroup.jobs),
                             "Jobs went into cooloff without the proper timing")
            for job in skimJobGroup.jobs:
                self.setJobTime.execute(jobID=job["id"],
                                        stateTime=int(time.time()) - 30 * pow(count, 2) - 5)
            testRetryManager.algorithm(None)

            idList = self.getJobs.execute(state='created')
            self.assertEqual(len(idList), len(skimJobGroup.jobs),
                             "Jobs didn't change state correctly")

    def testY_MultipleIterations(self):
        """
        _MultipleIterations_

        Paranoia based check to see if I'm saving class instances correctly
        """

        testJobGroup = self.createTestJobGroup(nJobs=self.nJobs)

        config = self.getConfig()
        changer = ChangeState(config)
        changer.propagate(testJobGroup.jobs, 'submitfailed', 'Created')
        changer.propagate(testJobGroup.jobs, 'submitcooloff', 'submitfailed')

        idList = self.getJobs.execute(state='SubmitCooloff')
        self.assertEqual(len(idList), self.nJobs)

        testRetryManager = RetryManagerPoller(config)
        testRetryManager.setup(None)

        for job in testJobGroup.jobs:
            self.setJobTime.execute(jobID=job["id"],
                                    stateTime=int(time.time()) - 50)

        testRetryManager.algorithm(None)
        idList = self.getJobs.execute(state='SubmitCooloff')
        self.assertEqual(len(idList), self.nJobs)

        for job in testJobGroup.jobs:
            self.setJobTime.execute(jobID=job["id"],
                                    stateTime=int(time.time()) - 150)

        testRetryManager.algorithm(None)

        idList = self.getJobs.execute(state='SubmitCooloff')
        self.assertEqual(len(idList), 0)

        idList = self.getJobs.execute(state='Created')
        self.assertEqual(len(idList), self.nJobs)

        # Make a new jobGroup for a second run
        testJobGroup = self.createTestJobGroup(nJobs=self.nJobs)

        # Set job state
        changer.propagate(testJobGroup.jobs, 'submitfailed', 'created')
        changer.propagate(testJobGroup.jobs, 'submitcooloff', 'submitfailed')

        # Set them to go off
        for job in testJobGroup.jobs:
            self.setJobTime.execute(jobID=job["id"],
                                    stateTime=int(time.time()) - 200)

        testRetryManager.algorithm(None)

        idList = self.getJobs.execute(state='SubmitCooloff')
        self.assertEqual(len(idList), 0)

        idList = self.getJobs.execute(state='Created')
        self.assertEqual(len(idList), self.nJobs * 2)

        return

    @attr('integration')
    def testZ_Profile(self):
        """
        _Profile_

        Do a basic profiling of the algo
        """

        import pstats

        nJobs = 1000

        testJobGroup = self.createTestJobGroup(nJobs=nJobs)

        config = self.getConfig()
        changer = ChangeState(config)
        changer.propagate(testJobGroup.jobs, 'createfailed', 'new')
        changer.propagate(testJobGroup.jobs, 'createcooloff', 'createfailed')

        idList = self.getJobs.execute(state='CreateCooloff')
        self.assertEqual(len(idList), nJobs)

        testRetryManager = RetryManagerPoller(config)
        testRetryManager.setup(None)

        for job in testJobGroup.jobs:
            self.setJobTime.execute(jobID=job["id"],
                                    stateTime=int(time.time()) - 50)

        testRetryManager.algorithm(None)
        idList = self.getJobs.execute(state='CreateCooloff')
        self.assertEqual(len(idList), nJobs)

        for job in testJobGroup.jobs:
            self.setJobTime.execute(jobID=job["id"],
                                    stateTime=int(time.time()) - 150)

        startTime = time.time()
        # cProfile.runctx("testRetryManager.algorithm()", globals(), locals(), filename = "profStats.stat")
        testRetryManager.algorithm(None)
        stopTime = time.time()

        idList = self.getJobs.execute(state='CreateCooloff')
        self.assertEqual(len(idList), 0)

        idList = self.getJobs.execute(state='New')
        self.assertEqual(len(idList), nJobs)

        print("Took %f seconds to run polling algo" % (stopTime - startTime))

        p = pstats.Stats('profStats.stat')
        p.sort_stats('cumulative')
        p.print_stats(0.2)

        return


if __name__ == '__main__':
    unittest.main()
